from os.path import dirname, join, realpath

from clerkai.transactions.parsers.se.banknorwegian.xlsx import \
    banknorwegian_se_xlsx_transactions_parser

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_banknorwegian_se_xlsx_transactions_parser():
    # type: () -> None
    transaction_file_path = join(test_data_dir_path, "Statement aug norwegian.xlsx")
    transactions_df = banknorwegian_se_xlsx_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



import os

import geopy.distance
import pandas as pd
import reverse_geocoder as rg

from clerkai.utils import list_files_in_clerk_input_subfolder


def location_history_flow(
    location_history_files_editable_columns,
    location_history_by_date_editable_columns,
    clerkai_input_folder_path,
    possibly_edited_df,
    location_history_folder_path,
    acknowledge_changes_in_clerkai_input_folder,
    current_history_reference,
    keep_unmerged_previous_edits=False,
    failfast=False,
):
    def list_location_history_files_in_location_history_folder():
        _ = list_files_in_clerk_input_subfolder(
            location_history_folder_path,
            clerkai_input_folder_path=clerkai_input_folder_path,
        )
        if len(_) == 0:
            return _
        for column in location_history_files_editable_columns:
            _[column] = None
        _["History reference"] = current_history_reference()
        return _[
            [
                "File name",
                "File path",
                *location_history_files_editable_columns,
                "File metadata",
                "History reference",
            ]
        ]

    location_history_files_df = list_location_history_files_in_location_history_folder()
    record_type = "location_history_files"
    location_history_files_first_columns = [
        "File name",
        "File path",
        *location_history_files_editable_columns,
    ]
    location_history_files_export_columns = [
        *location_history_files_first_columns,
        *location_history_files_df.columns.difference(
            location_history_files_first_columns
        ),
    ]
    # print("location_history_files_export_columns", location_history_files_export_columns)
    location_history_files_export_df = location_history_files_df.reindex(
        location_history_files_export_columns, axis=1
    )

    # Todo
    """
    def guess_content_type_based_on_filename():
        pass
    """
    # location_history_files_export_df[""]

    possibly_edited_location_history_files_df = possibly_edited_df(
        location_history_files_export_df,
        record_type,
        location_history_files_editable_columns,
        keep_unmerged_previous_edits,
    )

    included_location_history_files = possibly_edited_location_history_files_df[
        possibly_edited_location_history_files_df["Ignore"] != 1
    ]

    # make sure that the edited column values yields new commits
    # so that edit-files are dependent on the editable columns of file metadata
    location_history_files_editable_data_df = included_location_history_files[
        location_history_files_editable_columns + ["File metadata"]
    ]
    save_location_history_files_editable_data_in_location_history_folder(
        location_history_folder_path, location_history_files_editable_data_df
    )
    acknowledge_changes_in_clerkai_input_folder()

    from clerkai.location_history.parse import parse_location_history_files

    parsed_location_history_files = parse_location_history_files(
        included_location_history_files, clerkai_input_folder_path, failfast
    )

    unsuccessfully_parsed_location_history_files = parsed_location_history_files[
        ~parsed_location_history_files["Error"].isnull()
    ].drop(["File metadata", "Parse results"], axis=1)

    successfully_parsed_location_history_files = parsed_location_history_files[
        parsed_location_history_files["Error"].isnull()
    ].drop(["Error"], axis=1)

    if len(successfully_parsed_location_history_files) > 0:
        # concat all location_history
        all_parsed_location_history_df = pd.concat(
            successfully_parsed_location_history_files["Parse results"].values,
            sort=False,
        ).reset_index(drop=True)
        all_parsed_location_history_df[
            "History reference"
        ] = current_history_reference()
        # include location_history_files data
        all_parsed_location_history_df = pd.merge(
            all_parsed_location_history_df,
            included_location_history_files.drop(["Ignore"], axis=1).add_prefix(
                "Source location history file: "
            ),
            left_on="Source location history file index",
            right_index=True,
            suffixes=(False, False),
        )
        # print("all_parsed_location_history_df.columns", all_parsed_location_history_df.columns)
        have_timestamp_and_coordinates_mask = (
            ~all_parsed_location_history_df["Timestamp"].isnull()
            & ~all_parsed_location_history_df["Latitude"].isnull()
            & ~all_parsed_location_history_df["Longitude"].isnull()
        )
        location_history_df = all_parsed_location_history_df[
            have_timestamp_and_coordinates_mask
        ].sort_values(
            ["Timestamp"]
        )  # .drop_duplicates(subset=["ID"])

        # TODO: only in-memory location history, then editable day-by-day summary

        # reverse geo-coding
        location_history_df["coordinates"] = location_history_df.apply(
            lambda df: (df["Latitude"], df["Longitude"]), axis=1
        )
        unique_coordinates = list(set(list(location_history_df["coordinates"])))

        results = rg.search(unique_coordinates)  # default mode = 2
        reverse_geocoded_coordinates_df = pd.DataFrame(results)
        reverse_geocoded_coordinates_df["coordinates"] = unique_coordinates

        location_history_with_geonames_df = location_history_df.merge(
            reverse_geocoded_coordinates_df, on="coordinates"
        )

        location_history_with_geonames_df["date"] = location_history_with_geonames_df[
            "Timestamp"
        ].apply(lambda ts: ts.strftime("%Y-%m-%d"))
        location_history_with_geonames_df[
            "datehour"
        ] = location_history_with_geonames_df["Timestamp"].apply(
            lambda ts: ts.strftime("%Y-%m-%d %H")
        )
        groupby_date = location_history_with_geonames_df.groupby(["date"])
        groupby_datehour = location_history_with_geonames_df.groupby(["datehour"])

        def drop_duplicates_retain_order(a_list):
            return pd.Series(a_list).drop_duplicates().tolist()

        def geo_group_attributes(group):
            df = group.size().reset_index(name="counts")
            df["different_geo_names"] = (
                group["name"]
                .apply(drop_duplicates_retain_order)
                .reset_index(name="different_geo_names")["different_geo_names"]
            )
            df["different_geo_admin1"] = (
                group["admin1"]
                .apply(drop_duplicates_retain_order)
                .reset_index(name="different_geo_admin1")["different_geo_admin1"]
            )
            df["different_geo_admin2"] = (
                group["admin2"]
                .apply(drop_duplicates_retain_order)
                .reset_index(name="different_geo_admin2")["different_geo_admin2"]
            )
            df["different_countries"] = (
                group["cc"]
                .apply(drop_duplicates_retain_order)
                .reset_index(name="different_countries")["different_countries"]
            )
            df["coordinates"] = (
                group["coordinates"]
                .apply(drop_duplicates_retain_order)
                .reset_index(name="coordinates")["coordinates"]
            )
            df["last_known_coordinates"] = df["coordinates"].shift(1)
            df["last_known_different_geo_names"] = df["different_geo_names"].shift(1)
            df["last_known_different_geo_admin1"] = df["different_geo_admin1"].shift(1)
            df["last_known_different_geo_admin2"] = df["different_geo_admin2"].shift(1)
            df["last_known_different_countries"] = df["different_countries"].shift(1)
            return df

        location_history_by_date_df = geo_group_attributes(groupby_date)
        location_history_by_date_df["last_known_date"] = location_history_by_date_df[
            "date"
        ].shift(1)
        location_history_by_datehour_df = geo_group_attributes(groupby_datehour)
        location_history_by_datehour_df[
            "last_known_datehour"
        ] = location_history_by_datehour_df["datehour"].shift(1)

        def km_distance_since_last(row):
            # print(row["coordinates"], row["last_known_coordinates"])
            if row["last_known_coordinates"] != row["last_known_coordinates"]:
                return None
            return geopy.distance.distance(
                row["coordinates"][-1], row["last_known_coordinates"][-1]
            ).km

        location_history_by_date_df[
            "km_distance_since_last"
        ] = location_history_by_date_df.apply(km_distance_since_last, axis=1)
        location_history_by_datehour_df[
            "km_distance_since_last"
        ] = location_history_by_datehour_df.apply(km_distance_since_last, axis=1)

        # export location_history_by_date to xlsx
        record_type = "location_history_by_date"

        location_history_by_date_first_columns = [
            "last_known_date",
            "date",
            "km_distance_since_last",
            *location_history_by_date_editable_columns,
            "last_known_different_geo_names",
            "last_known_different_geo_admin1",
            "last_known_different_geo_admin2",
            "last_known_different_countries",
            "different_geo_names",
            "different_geo_admin1",
            "different_geo_admin2",
            "different_countries",
        ]
        location_history_by_date_export_columns = [
            *location_history_by_date_first_columns,
            *location_history_by_date_df.columns.difference(
                location_history_by_date_first_columns, sort=False
            ),
        ]
        # print("location_history_by_date_export_columns", location_history_by_date_export_columns)
        location_history_by_date_export_df = location_history_by_date_df.reindex(
            location_history_by_date_export_columns, axis=1
        )

        # index by date
        dates = pd.to_datetime(location_history_by_date_export_df["date"])
        location_history_by_date_export_df = location_history_by_date_export_df.set_index(
            dates
        )

        # include all dates, so that the edit file can be used to fill in locations for missing dates as well
        all_days = pd.date_range(
            location_history_by_date_export_df.index.min(),
            location_history_by_date_export_df.index.max(),
            freq="D",
        )
        location_history_by_date_export_df = location_history_by_date_export_df.reindex(
            all_days
        )
        location_history_by_date_export_df["date"] = all_days

        possibly_edited_location_history_by_date_df = possibly_edited_df(
            location_history_by_date_export_df,
            record_type,
            location_history_by_date_editable_columns,
            keep_unmerged_previous_edits=False,
        )

        # index by date again after reading from edit file
        dates = pd.to_datetime(possibly_edited_location_history_by_date_df["date"])
        possibly_edited_location_history_by_date_df = possibly_edited_location_history_by_date_df.set_index(
            dates
        )

    else:
        all_parsed_location_history_df = []
        location_history_df = []
        location_history_with_geonames_df = []
        possibly_edited_location_history_by_date_df = []

    return (
        location_history_files_df,
        possibly_edited_location_history_files_df,
        unsuccessfully_parsed_location_history_files,
        successfully_parsed_location_history_files,
        all_parsed_location_history_df,
        location_history_df,
        location_history_with_geonames_df,
        possibly_edited_location_history_by_date_df,
    )


def save_location_history_files_editable_data_in_location_history_folder(
    location_history_folder_path, location_history_files_editable_data_df
):
    csv = location_history_files_editable_data_df.to_csv(index=False)
    file_path = os.path.join(
        location_history_folder_path, "location_history_files_editable_data.csv"
    )
    with open(file_path, "w") as f:
        f.write(csv)



from typing import List

default_time_tracking_files_editable_columns: List[str] = [
    "Content type",
    "Ignore",
    "Editable time tracking entries? (1/0)",
    "Client (Default)",
    "Project / Costpool / Theme (Default)",
    "Invoice (Default)",
]

default_tracked_time_by_file_editable_columns: List[str] = []

aggregation_level_editable_columns: List[str] = [
    "Client",
    "Project / Costpool / Theme",
    "Invoice",
    "Invoice row ordinal",
    "Invoice row item",
    "Invoice row item category",
    "Invoice row item reference",
    "Invoice hours",
]

default_tracked_time_by_file_and_date_editable_columns = (
    aggregation_level_editable_columns
)

default_tracked_time_by_file_date_and_annotation_editable_columns = (
    aggregation_level_editable_columns
)

default_tracked_time_by_file_datehour_and_annotation_editable_columns = (
    aggregation_level_editable_columns
)

default_time_tracking_entries_editable_columns: List[str] = [
    "Hours 1 (Edited)",
    "Annotation 1 (Edited)",
    "Hours 2 (Edited)",
    "Annotation 2 (Edited)",
    "Hours 3 (Edited)",
    "Annotation 3 (Edited)",
]



import pandas as pd
from pandas.core.frame import DataFrame

from clerkai.transactions.parsers.parse_utils import amount_to_rounded_decimal
from clerkai.utils import is_nan, ymd_date_to_naive_datetime_obj


def import_xolo_csv_transaction_file(transaction_file):
    # type: (str) -> DataFrame
    return pd.read_csv(transaction_file, dtype=str)


def xolo_expenses_vendor_category_bug_fixer(xolo_csv_record):
    if is_nan(xolo_csv_record["Category"]):
        return xolo_csv_record["Vendor"]
    # todo: make only remove from the end
    return xolo_csv_record["Vendor"].replace(xolo_csv_record["Category"], "")


def xolo_csv_transactions_to_general_clerk_format(df):
    # type: (DataFrame) -> DataFrame
    normalized_df = pd.DataFrame()
    normalized_df["Raw Real Date"] = df[
        "Invoice date"
    ]  # not perfect, but better than nothing
    normalized_df["Raw Bank Date"] = df["Paid date"]
    normalized_df["Raw Payee"] = df["Vendor"]
    normalized_df["Raw Bank Message"] = df["Description"]
    normalized_df["Raw Amount"] = df["Amount"]
    normalized_df["Raw Balance"] = None
    normalized_df["Raw Currency"] = df["Currency"]
    normalized_df["Raw Doc Status"] = df["Status"]
    normalized_df["Raw Payment Status"] = df["Status.1"]
    normalized_df["Real Date"] = normalized_df["Raw Real Date"].apply(
        ymd_date_to_naive_datetime_obj
    )
    normalized_df["Bank Date"] = normalized_df["Raw Bank Date"].apply(
        ymd_date_to_naive_datetime_obj
    )
    normalized_df["Payee"] = df.apply(xolo_expenses_vendor_category_bug_fixer, axis=1)
    normalized_df["Bank Message"] = normalized_df["Raw Bank Message"]
    normalized_df["Amount"] = normalized_df["Raw Amount"].apply(
        amount_to_rounded_decimal
    )
    normalized_df["Balance"] = None
    normalized_df["Currency"] = normalized_df["Raw Currency"]
    normalized_df["Doc Status"] = normalized_df["Raw Doc Status"]
    normalized_df["Payment Status"] = normalized_df["Raw Payment Status"]
    normalized_df["Original data"] = df[
        [
            "Vendor",
            "Category",
            "Description",
            "Status",
            "Invoice date",
            "Paid date",
            "Source",
            "Amount",
            "Currency",
            "Status.1",
        ]
    ].to_dict(orient="records")
    return normalized_df[
        [
            "Real Date",
            "Bank Date",
            "Payee",
            "Bank Message",
            "Amount",
            "Balance",
            "Currency",
            "Doc Status",
            "Payment Status",
            "Original data",
            "Raw Real Date",
            "Raw Bank Date",
            "Raw Payee",
            "Raw Bank Message",
            "Raw Amount",
            "Raw Balance",
            "Raw Currency",
            "Raw Doc Status",
            "Raw Payment Status",
        ]
    ]


def xolo_csv_transactions_parser(transaction_file):
    # type: (str) -> DataFrame
    df = import_xolo_csv_transaction_file(transaction_file)
    # print(df.dtypes, df.columns)
    return xolo_csv_transactions_to_general_clerk_format(df)



import json

import pandas as pd
from joblib import Memory

from clerkai.time_tracking.parsers.neamtime.tslog import \
    neamtime_tslog_time_tracking_entries_parser
from clerkai.utils import clerkai_input_file_path, is_nan, raw_if_available

memory = Memory(location="/tmp", verbose=0)

parser_by_content_type = {
    "exported-time-tracking-file/neamtime-tslog": neamtime_tslog_time_tracking_entries_parser,
}


def naive_time_tracking_entry_ids(time_tracking_entries):
    import jellyfish

    def generate_naive_time_tracking_entry_id(time_tracking_entry):
        def none_if_nan(x):
            if is_nan(x):
                return None
            else:
                return x

        id_key_dict = {}
        id_key_dict["utc_timestamp"] = none_if_nan(
            raw_if_available("UTC Timestamp", time_tracking_entry)
        )
        source_lines_summary = none_if_nan(
            raw_if_available("Source Lines Summary", time_tracking_entry)
        )
        id_key_dict["source_lines_summary"] = (
            jellyfish.soundex(source_lines_summary)
            if type(source_lines_summary) is str
            else source_lines_summary
        )
        id_key_dict["session"] = none_if_nan(
            raw_if_available("Session", time_tracking_entry)
        )
        return json.dumps(id_key_dict, sort_keys=True, default=str, allow_nan=False)

    return time_tracking_entries.apply(generate_naive_time_tracking_entry_id, axis=1)


def naive_time_tracking_entry_id_duplicate_nums(time_tracking_entries):
    return (
        time_tracking_entries.groupby(["naive_time_tracking_entry_id"]).cumcount() + 1
    )


def time_tracking_entry_ids(time_tracking_entries):
    copy = time_tracking_entries.copy()
    copy["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(copy)
    copy[
        "naive_time_tracking_entry_id_duplicate_num"
    ] = naive_time_tracking_entry_id_duplicate_nums(copy)

    def generate_time_tracking_entry_id(time_tracking_entry):
        id_key_dict = {
            "ref": json.loads(time_tracking_entry["naive_time_tracking_entry_id"]),
            "ord": time_tracking_entry["naive_time_tracking_entry_id_duplicate_num"],
        }
        return json.dumps(id_key_dict)

    return copy.apply(generate_time_tracking_entry_id, axis=1)


def parse_time_tracking_files(
    time_tracking_files, clerkai_input_folder_path, keepraw=False, failfast=False
):
    class ContentTypeNotSetError(Exception):
        pass

    class ParserNotAvailableError(Exception):
        pass

    class ProcessingErrorsEncountered(Exception):
        pass

    class InvalidRowEncountered(Exception):
        pass

    def parse_time_tracking_file_row(time_tracking_file):
        results = {
            "time_tracking_entries": None,
            "parsing_metadata": None,
            "processing_errors": None,
        }
        error = None

        if failfast:
            print(
                '* Processing time tracking file with path: "%s/%s"'
                % (time_tracking_file["File path"], time_tracking_file["File name"]),
            )

        def parse():
            if is_nan(time_tracking_file["File path"]):
                # ignore invalid rows (eg unmerged edit rows without paths)
                error_msg = (
                    "Ignoring row %s since it has no file path"
                    % time_tracking_file.name
                )
                raise InvalidRowEncountered(error_msg)
            time_tracking_file_path = clerkai_input_file_path(
                clerkai_input_folder_path, time_tracking_file
            )
            content_type = time_tracking_file["Content type"]
            if not content_type:
                raise ContentTypeNotSetError(
                    "Time tracking file has no content type set"
                )
            if (
                content_type not in parser_by_content_type
                or not parser_by_content_type[content_type]
            ):
                raise ParserNotAvailableError(
                    "Content type '%s' has no parser" % content_type
                )
            parser = parser_by_content_type[content_type]
            # print(parser, time_tracking_file_path)

            @memory.cache
            def parse_file(_file_metadata, _content_type):
                return parser(time_tracking_file_path)

            time_tracking_entries, parsing_metadata, processing_errors = parse_file(
                time_tracking_file["File metadata"], content_type
            )

            # mark which source file was parsed, for later merging of the data
            time_tracking_entries[
                "Source time tracking file index"
            ] = time_tracking_file.name
            parsing_metadata[
                "Source time tracking file index"
            ] = time_tracking_file.name
            processing_errors[
                "Source time tracking file index"
            ] = time_tracking_file.name

            # time tracking entries ids
            if len(time_tracking_entries) > 0:
                # add future join/merge index
                time_tracking_entries["ID"] = time_tracking_entry_ids(
                    time_tracking_entries
                )
            else:
                time_tracking_entries["ID"] = None

            # drop raw columns
            if not keepraw:
                time_tracking_entries = time_tracking_entries.drop(
                    [
                        "Raw Ignore",
                        "Raw Source Line Numbers",
                        "Raw Source Lines Summary",
                        "Raw Session",
                        "Raw UTC Timestamp",
                        "Raw Work Date",
                        "Raw Date Before Parsing",
                        "Raw Timezone",
                        "Raw Time-annotated Source Lines Summary",
                        "Raw Hours",
                        "Raw Hours Rounded",
                        "Raw Annotation",
                        "Raw Year",
                        "Raw Year-half",
                        "Raw Month",
                        "Raw Week",
                        "Raw Client",
                        "Raw Invoice",
                        "Raw Invoice row ordinal",
                        "Raw Invoice row item",
                        "Raw Invoice row item category",
                        "Raw Invoice row item reference",
                        "Raw Invoice hours",
                    ],
                    axis=1,
                    errors="ignore",
                )
            return {
                "time_tracking_entries": time_tracking_entries,
                "parsing_metadata": parsing_metadata,
                "processing_errors": processing_errors,
            }

        # failfast raises errors except expected/benign value errors
        if failfast:
            try:
                results = parse()
            except (
                ContentTypeNotSetError,
                ParserNotAvailableError,
                InvalidRowEncountered,
            ) as e:
                error = e
        else:
            try:
                results = parse()
            except Exception as e:
                error = e

        # check for processing errors stemming from invalid/non-understood
        # time tracking files (rather than errors in the parser)
        if (
            results["processing_errors"] is not None
            and len(results["processing_errors"]) > 0
        ):
            error = ProcessingErrorsEncountered("Processing errors encountered")

        return pd.Series(
            [
                results["time_tracking_entries"],
                results["parsing_metadata"],
                results["processing_errors"],
                error,
            ],
            index=[
                "Parsed time tracking entries",
                "Parsing metadata",
                "Processing errors",
                "Error",
            ],
        )

    if len(time_tracking_files) == 0:
        raise Exception("No time tracking files to parse")

    parsed_time_tracking_file_results = time_tracking_files.apply(
        parse_time_tracking_file_row, axis=1
    )

    parsed_time_tracking_files = time_tracking_files.join(
        parsed_time_tracking_file_results
    )
    return parsed_time_tracking_files


def time_tracking_entries_from_parsed_time_tracking_files(parsed_time_tracking_files):
    time_tracking_entries_df = pd.concat(
        parsed_time_tracking_files["Parse results"].values, sort=False
    ).reset_index(drop=True)
    time_tracking_entries_df = pd.merge(
        time_tracking_entries_df,
        parsed_time_tracking_files.drop(
            ["Parse results", "History reference"], axis=1
        ).add_prefix("Source time tracking file: "),
        left_on="Source time tracking file index",
        right_index=True,
    )
    return time_tracking_entries_df.drop(["Source time tracking file index"], axis=1)



from typing import List

location_history_files_editable_columns: List[str] = [
    "Ignore",
    "Location history provider",
    "Content type",
]

location_history_by_date_editable_columns: List[str] = ["Trip"]

"""
travel_reports_editable_columns = [
    "Include in expense report",
    "Expense report receiver",
    "Expense report",
    "Expense report accounting period",
    "Destination",
    "Travel purpose",
    "Left date & time",
    "Returned date & time",
    "Other comment",
    "Days Of Personal Vacation",
    "Private Car Kilometers",
    "Days With Paid Breakfast",
    "Days With Paid Lunch",
    "Days With Paid Dinner",
]
"""



import hashlib
import os
from datetime import datetime
from os.path import getsize, join

import pandas as pd
from gspread import SpreadsheetNotFound, WorksheetNotFound
from gspread_dataframe import get_as_dataframe, set_with_dataframe
from gspread_formatting import CellFormat, Color
from gspread_formatting.dataframe import BasicFormatter, format_with_dataframe


def ensure_clerkai_folder_versioning(clerkai_input_folder_path):
    from git import Repo

    if os.path.isdir(os.path.join(clerkai_input_folder_path, ".git")):
        # TODO make sure that this is a clerk-managed git repository
        repo = Repo(clerkai_input_folder_path)
    else:
        repo = Repo.init(clerkai_input_folder_path)
        # make the commits in this repo be from clerk automation by default
        config = repo.config_writer()
        config.set_value("user", "name", "Clerk.ai")
        config.set_value("user", "email", "automation@clerk.ai")
        config.release()
        # initial (empty) commit
        repo.index.commit(message="Initial commit")
    assert not repo.bare
    return repo


# add all untracked and changed files
def add_all_untracked_and_changed_files(repo):
    # track ignores within the git folder
    with open(
        os.path.join(repo.working_tree_dir, ".git", "info", "exclude"), "w"
    ) as text_file:
        text_file.write("Edits\n.~lock.*\n")
    repo.git.add("-A")
    uncommited_changes = repo.git.status("--porcelain")
    if uncommited_changes != "":
        repo.git.commit("-m", "Current files", "-a")


def short_gitsha1(repo, sha):
    short_sha = repo.git.rev_parse(sha, short=1)
    return short_sha


def current_gitsha1(repo):
    sha = repo.head.object.hexsha
    return short_gitsha1(repo, sha)


def current_gitcommit_datetime(repo):
    from datetime import datetime

    return datetime.fromtimestamp(repo.head.commit.authored_date)


def commits_by_short_gitsha1(repo_path, repo):
    from pydriller import RepositoryMining

    commits_iterator = RepositoryMining(repo_path).traverse_commits()
    commits = {}
    for commit in commits_iterator:
        history_reference = short_gitsha1(repo, commit.hash)
        commits[history_reference] = commit

    if len(commits) == 0:
        raise Exception('No commits found in "%s"' % repo_path)

    return commits


def commit_datetime_from_history_reference(history_reference, commits):
    first_matching_commit_history_reference_key = next(
        filter(lambda _: _.startswith(history_reference), commits.keys()), False
    )
    if not first_matching_commit_history_reference_key:
        raise Exception('No commit matching "%s" found' % history_reference)
    return commits[first_matching_commit_history_reference_key].author_date


def clerkai_input_file_path(clerkai_input_folder_path, file):
    return os.path.join(
        file["File path"].replace("@", clerkai_input_folder_path), file["File name"]
    )


def export_file_name_by_record_type(record_type, suffix=""):
    file_extension = "xlsx"
    if record_type == "transaction_files":
        export_file_name_base = "Transaction files"
    elif record_type == "transactions":
        export_file_name_base = "Transactions"
    elif record_type == "receipt_files":
        export_file_name_base = "Receipt files"
    elif record_type == "location_history_files":
        export_file_name_base = "Location history files"
    elif record_type == "location_history_by_date":
        export_file_name_base = "Location history day-by-day"
    elif record_type == "time_tracking_files":
        export_file_name_base = "Time tracking files"
    elif record_type == "time_tracking_entries":
        export_file_name_base = "Time tracking entries"
    else:
        raise ValueError("record_type '%s' not recognized" % record_type)
    export_file_name = "%s%s.%s" % (export_file_name_base, suffix, file_extension)
    return (export_file_name, export_file_name_base)


class InvalidPreviouslyEditedDfException(BaseException):
    pass


def possibly_edited_df_util(
    current_commit_df,
    record_type,
    editable_columns,
    keep_unmerged_previous_edits,
    list_edit_files_in_edits_folder,
    current_history_reference,
    edits_folder_path,
    clerkai_input_folder_path,
    clerkai_input_folder_repo,
):
    # set config based on record type
    (export_file_name, export_file_name_base) = export_file_name_by_record_type(
        record_type
    )

    # list of edit files
    edit_files_df = list_edit_files_in_edits_folder()
    # print("edit_files_df", edit_files_df)

    # not much to do here if there are no edit files
    if len(edit_files_df) == 0:
        # make sure that the merged editable df file is available in the most current location
        return possibly_edited_commit_specific_df(
            df=current_commit_df,
            record_type=record_type,
            export_file_name=export_file_name,
            edits_folder_path=edits_folder_path,
            commit_datetime=current_gitcommit_datetime(clerkai_input_folder_repo),
            history_reference=current_history_reference(),
            create_if_not_exists=True,
        )

    # include earlier edits
    previous_main_edit_files_mask = (edit_files_df["File name"] == export_file_name) & (
        edit_files_df["Related history reference"] != current_history_reference()
    )

    # include gsheets edits
    gsheets_edit_files_mask = edit_files_df["File name"].str.contains(
        "%s.gsheets." % export_file_name_base
    )

    unmerged_non_current_main_edit_files = edit_files_df[
        previous_main_edit_files_mask | gsheets_edit_files_mask
    ]
    # print("unmerged_non_current_main_edit_files", unmerged_non_current_main_edit_files)

    # check if edit for the head commit already exists
    main_edit_file_df = possibly_edited_commit_specific_df(
        df=None,
        record_type=record_type,
        export_file_name=export_file_name,
        edits_folder_path=edits_folder_path,
        commit_datetime=current_gitcommit_datetime(clerkai_input_folder_repo),
        history_reference=current_history_reference(),
        create_if_not_exists=False,
    )
    main_edit_file_for_the_head_commit_exists = type(main_edit_file_df) is not bool

    # if edit for the head commit already exists and no other edit files are available - use it
    if (
        main_edit_file_for_the_head_commit_exists
        and len(unmerged_non_current_main_edit_files) == 0
    ):
        print(
            "Returning existing %s.xlsx (ignoring currently parsed data)"
            % (export_file_name_base)
        )
        # just one adjustment: make sure the currently configured editable
        # columns are available on the returned dataframe (despite them not being in the xlsx)
        for editable_column in editable_columns:
            if editable_column not in main_edit_file_df:
                main_edit_file_df[editable_column] = None

        return main_edit_file_df

    # include the current main edit file df if exists and a merge is
    # imminent - or else all changes only in the main edit file will be lost
    if main_edit_file_for_the_head_commit_exists:
        print(
            "Merging edits from %s edit file(s) and %s.xlsx into a new %s.xlsx (ignoring currently parsed data)"
            % (
                len(unmerged_non_current_main_edit_files),
                export_file_name_base,
                export_file_name_base,
            )
        )
        df_with_previous_edits_across_columns = main_edit_file_df
    else:
        print(
            "Merging edits from %s edit file(s) and the currently parsed data into %s.xlsx"
            % (len(unmerged_non_current_main_edit_files), export_file_name_base)
        )
        df_with_previous_edits_across_columns = current_commit_df

    columns_to_drop_after_propagation_of_previous_edits = []
    for index, edit_file in unmerged_non_current_main_edit_files.iterrows():
        previous_possibly_edited_df_xlsx_path = os.path.join(
            edit_file["File path"].replace("@/Edits", edits_folder_path),
            edit_file["File name"],
        )
        previous_possibly_edited_df = pd.read_excel(
            previous_possibly_edited_df_xlsx_path
        )
        try:
            (
                df_with_previous_edits_across_columns,
                additional_columns_to_drop_after_propagation_of_previous_edits,
            ) = merge_changes_from_previous_possibly_edited_df(
                accumulating_df=df_with_previous_edits_across_columns,
                previous_possibly_edited_df=previous_possibly_edited_df,
                edit_file=edit_file,
                record_type=record_type,
                clerkai_input_folder_path=clerkai_input_folder_path,
                current_history_reference=current_history_reference,
                keep_unmerged_previous_edits=keep_unmerged_previous_edits,
            )
        except InvalidPreviouslyEditedDfException as e:
            print(
                "Warning: Ignoring invalid previous_possibly_edited_df (%s)" % e,
                previous_possibly_edited_df,
            )
            continue
        columns_to_drop_after_propagation_of_previous_edits = [
            *columns_to_drop_after_propagation_of_previous_edits,
            *additional_columns_to_drop_after_propagation_of_previous_edits,
        ]

    df_with_previous_edits = propagate_previous_edits_from_across_columns(
        df_with_previous_edits_across_columns,
        unmerged_non_current_main_edit_files,
        editable_columns,
    )

    # clean up irrelevant old columns (should have been merged and propagated already)
    if not keep_unmerged_previous_edits:
        clean_df_with_previous_edits = df_with_previous_edits.drop(
            columns_to_drop_after_propagation_of_previous_edits, axis=1
        )
    else:
        print("Warning: Keeping potential old edits and columns for reference")
        clean_df_with_previous_edits = df_with_previous_edits

    def archive_edit_file(edit_file_to_archive):
        import shutil
        from datetime import datetime

        from_folder = edit_file_to_archive["File path"].replace(
            "@/Edits", edits_folder_path
        )
        to_folder = edit_file_to_archive["File path"].replace(
            "@/Edits", "@/Edits/Archive"
        ).replace("@/Edits", edits_folder_path) + (
            "/Archived %s" % datetime.today().strftime("%Y-%m-%d %H%M%S")
        )
        os.makedirs(to_folder, exist_ok=True)
        shutil.move(
            os.path.join(from_folder, edit_file_to_archive["File name"]),
            os.path.join(to_folder, edit_file_to_archive["File name"]),
        )

    # if the main edit file existed before the merging, archive it before
    main_edit_file_for_the_head_commit_mask = (
        edit_files_df["File name"] == export_file_name
    ) & (edit_files_df["Related history reference"] == current_history_reference())
    main_edit_file_for_the_head_commit = edit_files_df[
        main_edit_file_for_the_head_commit_mask
    ]
    main_edit_file_for_the_head_commit.apply(archive_edit_file, axis=1)

    # make sure that the merged editable df file is available in the most current location
    possibly_edited_df_with_previous_edits = possibly_edited_commit_specific_df(
        df=clean_df_with_previous_edits,
        record_type=record_type,
        export_file_name=export_file_name,
        edits_folder_path=edits_folder_path,
        commit_datetime=current_gitcommit_datetime(clerkai_input_folder_repo),
        history_reference=current_history_reference(),
        create_if_not_exists=True,
    )

    # at this point, we have incorporated the information from the edit files that were used here
    # thus, we move them to the archive folder
    unmerged_non_current_main_edit_files.apply(archive_edit_file, axis=1)

    return possibly_edited_df_with_previous_edits


def possibly_edited_commit_specific_df(
    df,
    record_type,
    export_file_name,
    edits_folder_path,
    commit_datetime,
    history_reference,
    create_if_not_exists,
    create_if_exists=False,
):
    (
        exists,
        commit_specific_directory,
        commit_specific_directory_path,
        xlsx_path,
    ) = edited_commit_specific_df_exists(
        export_file_name, edits_folder_path, commit_datetime, history_reference
    )
    if not exists and not create_if_not_exists:
        return False
    if create_if_exists or (not exists and create_if_not_exists):
        save_edited_commit_specific_df(
            df,
            commit_specific_directory,
            commit_specific_directory_path,
            export_file_name,
            record_type,
            xlsx_path,
        )
    return pd.read_excel(xlsx_path)


def edited_commit_specific_df_exists(
    unsanitized_export_file_name, edits_folder_path, commit_datetime, history_reference
):
    import pytz
    from pathvalidate import sanitize_filename

    utc_commit_datetime = commit_datetime.astimezone(pytz.utc)

    commit_specific_directory = "%s (%s)" % (
        utc_commit_datetime.strftime("%Y-%m-%d %H%M"),
        history_reference,
    )
    commit_specific_directory_path = os.path.join(
        edits_folder_path, commit_specific_directory
    )
    export_file_name = sanitize_filename(unsanitized_export_file_name)
    xlsx_path = os.path.join(commit_specific_directory_path, export_file_name)

    # print("Checking if '%s/%s' exists" % (commit_specific_directory, export_file_name))
    exists = os.path.isfile(xlsx_path)
    return exists, commit_specific_directory, commit_specific_directory_path, xlsx_path


def save_edited_commit_specific_df(
    df,
    commit_specific_directory,
    commit_specific_directory_path,
    export_file_name,
    record_type,
    xlsx_path,
):
    print("Creating '%s/%s'" % (commit_specific_directory, export_file_name))
    if not os.path.isdir(commit_specific_directory_path):
        os.mkdir(commit_specific_directory_path)
    export_columns = df.columns
    export_df = df[export_columns]
    with pd.ExcelWriter(xlsx_path, engine="xlsxwriter") as writer:
        if record_type == "transactions":
            export_transactions_xlsx(export_df, writer)
        else:
            export_df.to_excel(
                writer, sheet_name="Data", index=False, freeze_panes=(1, 0)
            )


def set_export_transactions_formulas(df, eu_locale=False):
    from xlsxwriter.utility import xl_col_to_name
    import numpy as np

    def col_letter(column_name):
        return xl_col_to_name(df.columns.get_loc(column_name))

    # set formulas
    df["Account"] = '=%s[ROW]&" - "&%s[ROW]' % (
        col_letter("Source transaction file: Account provider"),
        col_letter("Source transaction file: Account"),
    )
    if eu_locale:
        df["Date"] = '=IF(%s[ROW]<>"";%s[ROW];%s[ROW])' % (
            col_letter("Real Date"),
            col_letter("Real Date"),
            col_letter("Bank Date"),
        )
        df["Year"] = '=IF(%s[ROW]="";"";TEXT(%s[ROW]; "yyyy"))' % (
            col_letter("Date"),
            col_letter("Date"),
        )
        df["Month"] = '=IF(%s[ROW]="";"";TEXT(%s[ROW]; "yyyy-mm"))' % (
            col_letter("Date"),
            col_letter("Date"),
        )
    else:
        df["Date"] = '=IF(%s[ROW]<>"",%s[ROW],%s[ROW])' % (
            col_letter("Real Date"),
            col_letter("Real Date"),
            col_letter("Bank Date"),
        )
        df["Year"] = '=IF(%s[ROW]="","",TEXT(%s[ROW], "yyyy"))' % (
            col_letter("Date"),
            col_letter("Date"),
        )
        df["Month"] = '=IF(%s[ROW]="","",TEXT(%s[ROW], "yyyy-mm"))' % (
            col_letter("Date"),
            col_letter("Date"),
        )

    df["Row number at export"] = np.arange(len(df)) + 2

    # make formulas row-specific
    def insert_row_numbers(df_row, colname):
        return df_row[colname].replace("[ROW]", str(df_row["Row number at export"]))

    df["Account"] = df.apply(insert_row_numbers, axis=1, colname="Account")
    df["Date"] = df.apply(insert_row_numbers, axis=1, colname="Date")
    df["Year"] = df.apply(insert_row_numbers, axis=1, colname="Year")
    df["Month"] = df.apply(insert_row_numbers, axis=1, colname="Month")

    return df


def export_transactions_xlsx(export_df, writer):
    from xlsxwriter.utility import xl_col_to_name

    export_df = set_export_transactions_formulas(export_df)
    export_df.to_excel(writer, sheet_name="Data", index=False, freeze_panes=(1, 0))

    # adjust styles etc
    workbook = writer.book
    worksheet = workbook.get_worksheet_by_name("Data")
    # set default column width
    default_column_width = 10
    last_column_index = len(export_df.columns) - 1
    worksheet.set_column(
        "%s:%s" % (xl_col_to_name(0), xl_col_to_name(last_column_index),),
        default_column_width,
    )
    # account column
    account_column_index = export_df.columns.get_loc("Account")
    account_column_letter = xl_col_to_name(account_column_index)
    worksheet.set_column("%s:%s" % (account_column_letter, account_column_letter), 30)
    # date column
    date_column_index = export_df.columns.get_loc("Date")
    date_format = workbook.add_format({"num_format": "yyyy-mm-dd"})
    date_column_letter = xl_col_to_name(date_column_index)
    worksheet.set_column(
        "%s:%s" % (date_column_letter, date_column_letter), None, date_format
    )
    worksheet.set_column("%s:%s" % (date_column_letter, date_column_letter), 20)
    # TODO: possibly pre-calculate values of formulas to avoid LibreOffice issue
    # see https://stackoverflow.com/questions/32205927/xlsxwriter-and-libreoffice-not-showing-formulas-result


def export_to_gsheets(
    gsheets_client,
    export_df,
    gsheets_title,
    gsheets_sheet_name,
    record_type,
    create_if_not_exists=False,
    eu_locale=False,
    editable_columns=None,
):
    # open target gsheet
    try:
        sh = gsheets_client.open(gsheets_title)
    except SpreadsheetNotFound as e:
        if create_if_not_exists:
            sh = gsheets_client.create(gsheets_title)
        else:
            raise e

    try:
        worksheet = sh.worksheet(gsheets_sheet_name)
    except WorksheetNotFound as e:
        if create_if_not_exists:
            worksheet = sh.add_worksheet(title=gsheets_sheet_name, rows=1, cols=1)
        else:
            raise e

    # set what to export
    df = export_df.copy()

    if len(worksheet.row_values(1)) > 0:
        # get existing worksheet contents
        existing_df = get_as_dataframe(worksheet)
        if len(existing_df.columns) > 0:
            # re-index columns to match the existing data + add any new columns (since the latest export/edit round)
            columns_not_in_existing_df = export_df.columns.difference(
                existing_df.columns, sort=False
            )
            df = df.reindex(
                existing_df.columns.tolist() + columns_not_in_existing_df.tolist(),
                axis=1,
            )

    # set formulas for export
    if len(df) > 0:
        if (
            record_type == "transaction_files"
            or record_type == "receipt_files"
            or record_type == "location_history_files"
            or record_type == "time_tracking_files"
        ):
            pass
        elif record_type == "transactions":
            df = set_export_transactions_formulas(df, eu_locale)
        elif record_type == "location_history_by_date":
            pass
        elif record_type == "time_tracking_entries":
            pass
        else:
            raise ValueError("record_type '%s' not recognized" % record_type)

    # export to gsheets
    # if len(df) == 0:
    #     set_frozen(worksheet, rows=0)
    set_with_dataframe(worksheet, df, resize=True)
    # if len(df) > 0:
    #     set_frozen(worksheet, rows=1)

    dark_purple = Color(56 / 255, 40 / 255, 54 / 255)
    # orange = Color(252 / 255, 142 / 255, 30 / 255)
    white = Color(255 / 255, 255 / 255, 255 / 255)
    # light_orange_3 = Color(255 / 255, 229 / 255, 227 / 255)
    light_grey = Color(240 / 255, 240 / 255, 240 / 255)

    editable_column_cell_format = CellFormat(backgroundColor=white,)
    non_editable_column_cell_format = CellFormat(backgroundColor=light_grey,)

    column_formats = {}
    if editable_columns:
        non_editable_columns = df.columns.difference(editable_columns, sort=False)
        for non_editable_column_name in non_editable_columns:
            column_formats[non_editable_column_name] = non_editable_column_cell_format
        for editable_column_name in editable_columns:
            column_formats[editable_column_name] = editable_column_cell_format

    formatter = BasicFormatter(
        header_background_color=dark_purple,
        header_text_color=white,
        decimal_format="#,##0.00",
        date_format="YYYY-MM-DD",
        freeze_headers=True,
        column_formats=column_formats,
    )
    format_with_dataframe(
        worksheet, df, formatter, include_index=False, include_column_header=True
    )

    spreadsheet_url = "https://docs.google.com/spreadsheets/d/%s" % sh.id
    return spreadsheet_url


def fetch_gsheets_worksheet_as_df(gsheets_client, gsheets_title, gsheets_sheet_name):
    sh = gsheets_client.open(gsheets_title)
    # spreadsheet_url = "https://docs.google.com/spreadsheets/d/%s" % sh.id
    # print(spreadsheet_url)
    worksheet = sh.worksheet(gsheets_sheet_name)
    return get_as_dataframe(worksheet)


def changes_between_two_commits(repo_base_path, from_commit, to_commit):
    from pydriller import RepositoryMining

    commits_iterator = RepositoryMining(
        repo_base_path, filepath=".", from_commit=from_commit, to_commit=to_commit
    ).traverse_commits()

    # follows file renames (ignores deletes and additions, so some new paths may have been
    # subsequently deleted, and some old paths may have not existed in the from_commit
    old_to_new_paths = {}
    new_to_old_paths = {}
    old_now_deleted_paths = {}  # TODO
    old_non_existing_now_added_paths = {}  # TODO
    for commit in commits_iterator:
        for modification in commit.modifications:
            if (
                modification.old_path
                and modification.new_path
                and modification.old_path != modification.new_path
            ):
                if modification.new_path in new_to_old_paths:
                    old_to_new_paths[
                        new_to_old_paths[modification.new_path]
                    ] = modification.new_path
                    new_to_old_paths[modification.new_path] = new_to_old_paths[
                        modification.new_path
                    ]
                else:
                    old_to_new_paths[modification.old_path] = modification.new_path
                    new_to_old_paths[modification.new_path] = modification.old_path
    del new_to_old_paths
    # print("old_to_new_paths", old_to_new_paths)

    return (old_to_new_paths, old_now_deleted_paths, old_non_existing_now_added_paths)


def merge_changes_from_previous_possibly_edited_df(
    accumulating_df,
    previous_possibly_edited_df,
    edit_file,
    record_type,
    clerkai_input_folder_path,
    current_history_reference,
    keep_unmerged_previous_edits,
):

    if type(previous_possibly_edited_df) is bool and not previous_possibly_edited_df:
        print("Context: edit_file", edit_file)
        raise Exception(
            "There is no previous possibly edited dataframe to merge changes from"
        )

    # set config based on record type
    if (
        record_type == "transaction_files"
        or record_type == "receipt_files"
        or record_type == "location_history_files"
        or record_type == "time_tracking_files"
    ):
        additional_join_column = None
        file_name_column_name = "File name"
        file_path_column_name = "File path"
    elif record_type == "transactions":
        additional_join_column = "ID"
        file_name_column_name = "Source transaction file: File name"
        file_path_column_name = "Source transaction file: File path"
        # Add ID column if not present in previous edits file
        if additional_join_column not in previous_possibly_edited_df.columns:
            from clerkai.transactions.parse import transaction_ids

            previous_possibly_edited_df[additional_join_column] = transaction_ids(
                previous_possibly_edited_df
            )
    elif record_type == "location_history_by_date":
        # for now simply return the current df, ignoring the previously possibly edited df
        return (accumulating_df, [])
        """
        # TODO: Support editing location history
        additional_join_column = "ID"
        file_name_column_name = "Source location history file: File name"
        file_path_column_name = "Source location history file: File path"
        # Add ID column if not present in previous edits file
        if additional_join_column not in previous_possibly_edited_df.columns:
            from clerkai.location_history.parse import location_history_ids

            previous_possibly_edited_df[additional_join_column] = location_history_ids(
                previous_possibly_edited_df
            )
        """
    elif record_type == "time_tracking_entries":
        additional_join_column = "ID"
        file_name_column_name = "Source time tracking file: File name"
        file_path_column_name = "Source time tracking file: File path"
        # Add ID column if not present in previous edits file
        if additional_join_column not in previous_possibly_edited_df.columns:
            from clerkai.transactions.parse import transaction_ids

            previous_possibly_edited_df[additional_join_column] = transaction_ids(
                previous_possibly_edited_df
            )
    else:
        raise ValueError("record_type '%s' not recognized" % record_type)

    # sanity check
    if file_name_column_name not in previous_possibly_edited_df.columns:
        raise InvalidPreviouslyEditedDfException(
            "Missing column: %s" % file_name_column_name
        )
    if file_path_column_name not in previous_possibly_edited_df.columns:
        raise InvalidPreviouslyEditedDfException(
            "Missing column: %s" % file_path_column_name
        )
    if additional_join_column not in previous_possibly_edited_df.columns:
        raise InvalidPreviouslyEditedDfException(
            "Missing column: %s" % additional_join_column
        )

    # print("df.head(), edit_file, previous_possibly_edited_df.head()")
    # print(df.head(), edit_file, previous_possibly_edited_df.head())

    # get relevant from and to commits between the old and new edit files
    from_commit = edit_file["Related history reference"]
    to_commit = current_history_reference()

    def normalize_encoding(string):
        import unicodedata as ud

        normalized = ud.normalize("NFC", string)
        return normalized

    def joined_normalized_path(record):
        # normalize encodings to properly join paths that have been encoded differently for whatever reason
        joined_path = "%s/%s" % (
            record[file_path_column_name],
            record[file_name_column_name],
        )
        return normalize_encoding(joined_path)

    accumulating_df["clerkai_path"] = accumulating_df.apply(
        joined_normalized_path, axis=1
    )
    left_on = ["clerkai_path"]

    if from_commit != to_commit:
        (
            old_to_new_paths,
            old_now_deleted_paths,
            old_non_existing_now_added_paths,
        ) = changes_between_two_commits(
            clerkai_input_folder_path, from_commit, to_commit
        )

        previous_possibly_edited_df["clerkai_path"] = previous_possibly_edited_df.apply(
            joined_normalized_path, axis=1
        )

        def find_head_commit_corresponding_clerkai_path(clerkai_path):
            clerkai_path_key = clerkai_path.replace("@/", "")
            if clerkai_path_key in old_to_new_paths:
                return "@/%s" % old_to_new_paths[clerkai_path_key]
            else:
                # if no moves occurred just use the old path as is
                return clerkai_path

        previous_possibly_edited_df[
            "head_commit_corresponding_clerkai_path"
        ] = previous_possibly_edited_df["clerkai_path"].apply(
            find_head_commit_corresponding_clerkai_path
        )

        right_on = ["head_commit_corresponding_clerkai_path"]
    else:
        previous_possibly_edited_df["clerkai_path"] = previous_possibly_edited_df.apply(
            joined_normalized_path, axis=1
        )
        right_on = ["clerkai_path"]

    suffix = " (%s - %s)" % (from_commit, edit_file["File name"])

    def add_suffix(column_name):
        return "%s%s" % (column_name, suffix)

    if additional_join_column:
        left_on.append(additional_join_column)
        right_on.append(additional_join_column)

    suffixed_previous_possibly_edited_df = previous_possibly_edited_df.add_suffix(
        suffix
    )

    import pandas as pd

    suffixed_right_on = [add_suffix(column_name) for column_name in right_on]

    merged_possibly_edited_df = pd.merge(
        accumulating_df,
        suffixed_previous_possibly_edited_df,
        how="left" if not keep_unmerged_previous_edits else "outer",
        left_on=left_on,
        right_on=suffixed_right_on,
        suffixes=(False, False),
    )

    # drop temporary merge columns
    merged_possibly_edited_df = merged_possibly_edited_df.drop(["clerkai_path"], axis=1)

    # hint that more columns may be dropped upon successful propagation of previous edits
    columns_to_drop_after_propagation_of_previous_edits = (
        suffixed_previous_possibly_edited_df.columns
    )

    return (
        merged_possibly_edited_df,
        columns_to_drop_after_propagation_of_previous_edits,
    )


def set_where_nan():
    pass


def propagate_previous_edits_from_across_columns(
    df_with_previous_edits_across_columns, unmerged_edit_files, editable_columns
):

    for index, edit_file in unmerged_edit_files.iterrows():
        suffix = " (%s - %s)" % (
            edit_file["Related history reference"],
            edit_file["File name"],
        )
        for column_name in editable_columns:
            if column_name not in df_with_previous_edits_across_columns.columns:
                # the editable column has not been seen before, make sure to initiate it
                df_with_previous_edits_across_columns[column_name] = None
            suffixed_column_name = "%s%s" % (column_name, suffix)
            if (
                suffixed_column_name
                not in df_with_previous_edits_across_columns.columns
            ):
                pass
            else:
                if column_name in df_with_previous_edits_across_columns.columns:
                    df_where_column_is_null_mask = df_with_previous_edits_across_columns[
                        column_name
                    ].isnull()
                    df_with_previous_edits_across_columns.loc[
                        df_where_column_is_null_mask, column_name
                    ] = df_with_previous_edits_across_columns[suffixed_column_name]
                else:
                    df_with_previous_edits_across_columns[
                        column_name
                    ] = df_with_previous_edits_across_columns[suffixed_column_name]
        # print("df_with_previous_edits_across_columns.head()", df_with_previous_edits_across_columns.head())

    return df_with_previous_edits_across_columns


def is_ignored_file(filename):
    if ".DS_Store" in filename:
        return True
    if ".gitignore" in filename:
        return True
    if "Icon" in filename:
        if filename.encode("utf-8").hex() == "49636f6e0d":
            return True
    if ".~lock" in filename:
        return True


def sha256sum(filename):
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()


def sha1sum(filename):
    h = hashlib.sha1()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()


def list_files_in_folder(folder_path):
    def is_not_ignored_file(filename):
        return not is_ignored_file(filename)

    all_files = []
    for root, dirs, files in os.walk(folder_path):
        # print(root, "consumes", end=" ")
        # print(sum(getsize(join(root, name)) for name in files), end=" ")
        # print("bytes in", len(files), "non-directory files")
        if ".git" in dirs:
            dirs.remove(".git")  # don't visit .git directories
        files = filter(is_not_ignored_file, files)
        # print(image_files)
        for file in list(files):
            file_sha256sum = sha256sum(join(root, file))
            file_sha1sum = sha1sum(join(root, file))
            all_files.append(
                {
                    "File name": file,
                    "File path": root,
                    "File metadata": {
                        "size": getsize(join(root, file)),
                        "sha1sum": file_sha1sum,
                        "sha256sum": file_sha256sum,
                    },
                }
            )
    return all_files


def list_files_in_clerk_subfolder(folder_path, clerkai_folder_path):
    import pandas as pd

    _ = pd.DataFrame(list_files_in_folder(folder_path))
    if len(_) > 0:
        _["File path"] = _["File path"].apply(
            lambda root: root.replace(clerkai_folder_path, "@")
        )
        # ignore *_editable_data.csv
        _ = _[~_["File name"].str.contains("_editable_data.csv$", regex=True)]
        # ignore .~lock files
        _ = _[~_["File name"].str.contains(".~lock")]
    return _


def list_files_in_clerk_input_subfolder(folder_path, clerkai_input_folder_path):
    import pandas as pd

    _ = pd.DataFrame(list_files_in_folder(folder_path))
    if len(_) > 0:
        _["File path"] = _["File path"].apply(
            lambda root: root.replace(clerkai_input_folder_path, "@")
        )
        # ignore *_editable_data.csv
        _ = _[~_["File name"].str.contains("_editable_data.csv$", regex=True)]
        # ignore .~lock files
        _ = _[~_["File name"].str.contains(".~lock")]
    return _


def is_nan(x):
    import math

    try:
        return math.isnan(x)
    except TypeError:
        return False


def ymd_date_to_naive_datetime_obj(datetime_str):
    if is_nan(datetime_str):
        return None
    datetime_obj = datetime.strptime(datetime_str, "%Y-%m-%d")
    return datetime_obj


def fi_dmy_date_to_naive_datetime_obj(datetime_str):
    if is_nan(datetime_str):
        return None
    datetime_obj = datetime.strptime(datetime_str, "%d.%m.%Y")
    return datetime_obj


def raw_if_available(field_name, entry):
    raw_field_name = "Raw %s" % field_name
    if raw_field_name in entry and entry[raw_field_name] is not None:
        return entry[raw_field_name]
    if field_name in entry:
        return entry[field_name]
    else:
        return None



from decimal import Decimal
from io import StringIO

import pandas as pd

from clerkai.transactions.parsers.parse_utils import (
    amount_to_rounded_decimal, convert_european_amount_to_decimal)


def test_amount_to_rounded_decimal():
    # type: () -> None
    foo = amount_to_rounded_decimal("123.45")
    assert foo == Decimal("123.45")
    # foo = amount_to_rounded_decimal("123,45")
    # assert foo == Decimal("123.45")
    foo = amount_to_rounded_decimal(123.45)
    assert foo == Decimal("123.45")


def test_convert_european_amount_to_decimal():
    # type: () -> None
    # eu formats:
    foo = convert_european_amount_to_decimal("123,45")
    assert foo == Decimal("123.45")
    foo = convert_european_amount_to_decimal("1.123,45")
    assert foo == Decimal("1123.45")
    # not eu formats:
    foo = convert_european_amount_to_decimal("123.45")
    assert foo == Decimal("12345.00")
    foo = convert_european_amount_to_decimal(123.45)
    assert foo == Decimal("12345.00")


def test_read_csv_with_decimal():
    # type: () -> None
    df = pd.read_csv(
        StringIO("foo,bar,amount\nzoo,zar,123.45\n"),
        converters={"amount": convert_european_amount_to_decimal},
    )
    assert type(df["amount"][0]) == Decimal
    # df = pd.read_csv(StringIO("foo,bar,amount\nzoo,zar,123.45\n"), dtype={'amount': Decimal})
    # assert type(df["amount"][0]) == Decimal



import os

import pandas as pd

from clerkai.utils import list_files_in_clerk_input_subfolder


def transactions_flow(
    transaction_files_editable_columns,
    transactions_editable_columns,
    clerkai_input_folder_path,
    possibly_edited_df,
    transactions_folder_path,
    acknowledge_changes_in_clerkai_input_folder,
    current_history_reference,
    keep_unmerged_previous_edits=False,
    failfast=False,
):
    def list_transaction_files_in_transactions_folder():
        _ = list_files_in_clerk_input_subfolder(
            transactions_folder_path,
            clerkai_input_folder_path=clerkai_input_folder_path,
        )
        if len(_) == 0:
            return _
        for column in transaction_files_editable_columns:
            _[column] = None
        _["History reference"] = current_history_reference()
        return _[
            [
                "File name",
                "File path",
                *transaction_files_editable_columns,
                "File metadata",
                "History reference",
            ]
        ]

    transaction_files_df = list_transaction_files_in_transactions_folder()
    record_type = "transaction_files"
    transaction_files_first_columns = [
        "File name",
        "File path",
        *transaction_files_editable_columns,
    ]
    transaction_files_export_columns = [
        *transaction_files_first_columns,
        *transaction_files_df.columns.difference(transaction_files_first_columns),
    ]
    # print("transaction_files_export_columns", transaction_files_export_columns)
    transaction_files_export_df = transaction_files_df.reindex(
        transaction_files_export_columns, axis=1
    )

    # Todo
    """
    def guess_content_type_based_on_filename():
        pass
    """

    possibly_edited_transaction_files_df = possibly_edited_df(
        transaction_files_export_df,
        record_type,
        transaction_files_editable_columns,
        keep_unmerged_previous_edits,
    )

    included_transaction_files = possibly_edited_transaction_files_df[
        possibly_edited_transaction_files_df["Ignore"] != 1
    ]

    if len(included_transaction_files) > 0:

        # make sure that the edited column values yields new commits
        # so that edit-files are dependent on the editable columns of file metadata
        transaction_files_editable_data_df = included_transaction_files[
            transaction_files_editable_columns + ["File metadata"]
        ]
        save_transaction_files_editable_data_in_transactions_folder(
            transactions_folder_path, transaction_files_editable_data_df
        )
        acknowledge_changes_in_clerkai_input_folder()

        from clerkai.transactions.parse import parse_transaction_files

        parsed_transaction_files = parse_transaction_files(
            transaction_files=included_transaction_files,
            clerkai_input_folder_path=clerkai_input_folder_path,
            failfast=failfast,
        )

        unsuccessfully_parsed_transaction_files = parsed_transaction_files[
            ~parsed_transaction_files["Error"].isnull()
        ].drop(["File metadata", "Parse results"], axis=1)

        successfully_parsed_transaction_files = parsed_transaction_files[
            parsed_transaction_files["Error"].isnull()
        ].drop(["Error"], axis=1)

    else:
        unsuccessfully_parsed_transaction_files = []
        successfully_parsed_transaction_files = []

    if len(successfully_parsed_transaction_files) > 0:
        # concat all transactions
        all_parsed_transactions_df = pd.concat(
            successfully_parsed_transaction_files["Parse results"].values, sort=False
        ).reset_index(drop=True)
        all_parsed_transactions_df["History reference"] = current_history_reference()
        # include transaction_files data
        all_parsed_transactions_df = pd.merge(
            all_parsed_transactions_df,
            included_transaction_files.drop(["Ignore"], axis=1).add_prefix(
                "Source transaction file: "
            ),
            left_on="Source transaction file index",
            right_index=True,
            suffixes=(False, False),
        )

        # print("all_parsed_transactions_df.columns", all_parsed_transactions_df.columns)

        transactions_df = all_parsed_transactions_df.drop_duplicates(subset=["ID"])

        # ensure that empty currency values is filled with source file currency if available
        if "Currency" in transactions_df.columns:
            transactions_df_where_currency_column_is_null_mask = transactions_df[
                "Currency"
            ].isnull()
            transactions_df_where_source_transaction_file_account_currency_column_is_null_mask = transactions_df[
                "Source transaction file: Account currency"
            ].isnull()
            transactions_df.loc[
                transactions_df_where_currency_column_is_null_mask, "Currency",
            ] = transactions_df.loc[
                ~transactions_df_where_source_transaction_file_account_currency_column_is_null_mask,
                "Source transaction file: Account currency",
            ]
        else:
            transactions_df["Currency"] = transactions_df[
                "Source transaction file: Account currency"
            ]

        # export all transactions to xlsx
        record_type = "transactions"

        transactions_first_columns = [
            "Account",
            "Date",
            "Year",
            "Month",
            *transactions_editable_columns,
            "Real Date",
            "Bank Date",
        ]
        transactions_export_columns = [
            *transactions_first_columns,
            *transactions_df.columns.difference(transactions_first_columns, sort=False),
            "Row number at export",
        ]
        # print("transactions_export_columns", transactions_export_columns)
        transactions_export_df = transactions_df.reindex(
            transactions_export_columns, axis=1
        )

        # convert Decimal columns to float prior to export or excel will treat them as strings
        # todo: less hacky conversion of Decimal-columns
        from clerkai.utils import is_nan

        def float_if_not_nan(number):
            if is_nan(number) or number is None:
                return None
            return float(number)

        for column in [
            "Amount",
            "Balance",
            "Foreign Currency Rate",
            "Foreign Currency Amount",
        ]:
            if column in transactions_export_df.columns:
                transactions_export_df[column] = transactions_export_df[column].apply(
                    float_if_not_nan
                )

        possibly_edited_transactions_df = possibly_edited_df(
            transactions_export_df,
            record_type,
            transactions_editable_columns,
            keep_unmerged_previous_edits=keep_unmerged_previous_edits,
        )

    else:
        all_parsed_transactions_df = []
        transactions_df = []
        possibly_edited_transactions_df = []

    return (
        transaction_files_df,
        possibly_edited_transaction_files_df,
        unsuccessfully_parsed_transaction_files,
        successfully_parsed_transaction_files,
        all_parsed_transactions_df,
        transactions_df,
        possibly_edited_transactions_df,
        transaction_files_editable_columns,
        transactions_editable_columns,
    )


def save_transaction_files_editable_data_in_transactions_folder(
    transactions_folder_path, transaction_files_editable_data_df
):
    csv = transaction_files_editable_data_df.to_csv(index=False)
    file_path = os.path.join(
        transactions_folder_path, "transaction_files_editable_data.csv"
    )
    with open(file_path, "w") as f:
        f.write(csv)



from datetime import datetime

import pytz

from clerkai.utils import is_nan


def timestamp_ms_to_utc_datetime_obj(timestamp_ms):
    if is_nan(timestamp_ms):
        return None
    if type(timestamp_ms) is str:
        timestamp_ms = float(timestamp_ms)
    datetime_obj = datetime.fromtimestamp(timestamp_ms / 1000)
    return datetime_obj.astimezone(pytz.utc)


def exiftool_date_to_utc_datetime_obj(exiftool_date):
    if is_nan(exiftool_date):
        return None
    if exiftool_date == "0000:00:00 00:00:00":
        return None
    try:
        naive_parsed_datetime = datetime.strptime(exiftool_date, "%Y:%m:%d %H:%M:%S")
        parsed_datetime = pytz.utc.localize(naive_parsed_datetime)
    except ValueError:
        parsed_datetime = datetime.strptime(
            exiftool_date, "%Y:%m:%d %H:%M:%S%z"
        ).astimezone(pytz.utc)
    return parsed_datetime



from os.path import dirname, join, realpath

from clerkai.transactions.parsers.se.nordea.personal.internetbanken_privat.xlsx import (
    nordea_se_personal_internetbanken_privat_xlsx_transactions_parser,
    nordea_se_transaction_text_to_datetime_obj)

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_nordea_se_transaction_text_to_datetime_obj():
    assert (
        nordea_se_transaction_text_to_datetime_obj(
            "Kortköp 181008 Conrad Electronic"
        ).strftime("%Y-%m-%d")
        == "2018-10-08"
    )


def test_nordea_se_personal_internetbanken_privat_xlsx_transactions_parser():
    # type: () -> None
    transaction_file_path = join(test_data_dir_path, "export.personkonto.edited.xls")
    transactions_df = nordea_se_personal_internetbanken_privat_xlsx_transactions_parser(
        transaction_file_path
    )
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected


def test_nordea_se_personal_internetbanken_privat_xlsx_transactions_parser_2():
    # type: () -> None
    transaction_file_path = join(test_data_dir_path, "export.personkonto.edited.2.xls")
    transactions_df = nordea_se_personal_internetbanken_privat_xlsx_transactions_parser(
        transaction_file_path
    )
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



from clerkai.location_history.parsers.parse_utils import (
    exiftool_date_to_utc_datetime_obj, timestamp_ms_to_utc_datetime_obj)


def test_timestamp_ms_to_utc_datetime_obj():
    # type: () -> None
    # TODO: check if this is correct - the results should be the same regardless of the timezone of the os
    assert (
        timestamp_ms_to_utc_datetime_obj(0).strftime("%Y-%m-%d %H:%M%z")
        == "1970-01-01 00:00+0000"
    )
    assert (
        timestamp_ms_to_utc_datetime_obj(1384393776624).strftime("%Y-%m-%d %H:%M%z")
        == "2013-11-14 01:49+0000"
    )
    assert (
        timestamp_ms_to_utc_datetime_obj("1384393776624").strftime("%Y-%m-%d %H:%M%z")
        == "2013-11-14 01:49+0000"
    )
    assert (
        timestamp_ms_to_utc_datetime_obj("1574246049000").strftime("%Y-%m-%d %H:%M%z")
        == "2019-11-20 10:34+0000"
    )


def test_exiftool_date_to_utc_datetime_obj():
    # type: () -> None
    assert (
        exiftool_date_to_utc_datetime_obj("2017:08:26 22:09:57").strftime(
            "%Y-%m-%d %H:%M:%S%z"
        )
        == "2017-08-26 22:09:57+0000"
    )
    assert (
        exiftool_date_to_utc_datetime_obj("2017:08:26 22:09:57").strftime(
            "%Y-%m-%d %H:%M:%S%z"
        )
        == "2017-08-26 22:09:57+0000"
    )
    assert (
        exiftool_date_to_utc_datetime_obj("2017:08:26 22:09:57+02:00").strftime(
            "%Y-%m-%d %H:%M:%S%z"
        )
        == "2017-08-26 20:09:57+0000"
    )
    assert (
        exiftool_date_to_utc_datetime_obj("2017:08:26 22:09:57-08:00").strftime(
            "%Y-%m-%d %H:%M:%S%z"
        )
        == "2017-08-27 06:09:57+0000"
    )
    assert exiftool_date_to_utc_datetime_obj("0000:00:00 00:00:00") is None



from datetime import datetime
from io import StringIO

import pandas as pd
from pandas.core.frame import DataFrame

from clerkai.transactions.parsers.parse_utils import \
    convert_european_amount_to_decimal


def fi_date_to_datetime_obj(datetime_str):
    # type: (str) -> datetime
    datetime_obj = datetime.strptime(datetime_str, "%d.%m.%Y")
    return datetime_obj


def nordea_fi_reference_number_to_datetime_obj(reference_number):
    reference_number_str = str(reference_number)
    if not reference_number_str.isdigit():
        return None
    if not len(reference_number_str) == 12:
        return None
    datetime_str = reference_number_str[0:6]
    datetime_obj = datetime.strptime(datetime_str, "%y%m%d")
    return datetime_obj


def import_nordea_fi_lang_se_txt_transaction_file(transaction_file):
    # type: (str) -> DataFrame
    with open(transaction_file, "rb") as f:
        contents = f.read()
    # print(contents)
    decoded_contents = contents.decode("utf8")
    # originally, these files use "\n\r\n" as line separator (go figure)
    # but after editing in a sane editor, these will be "\n\n"
    # by replacing "\n\r\n" with "\n\n" we support both pristine and
    # edited files
    line_ending_normalized_contents = decoded_contents.replace("\n\r\n", "\n\n")
    header_split = line_ending_normalized_contents.split("\n\n", 1)
    # print(header_split)
    nordeafi_adjusted_contents = header_split[1].replace("\n\n", "\n")
    # print(nordeafi_adjusted_contents)
    return pd.read_csv(
        StringIO(nordeafi_adjusted_contents),
        sep="\t",
        thousands=".",
        decimal=",",
        dtype=str,
    )


def nordea_fi_lang_se_txt_transactions_to_general_clerk_format(df):
    # type: (DataFrame) -> DataFrame
    normalized_df = pd.DataFrame()
    normalized_df["Raw Real Date"] = df["Referens"]
    normalized_df["Raw Bank Date"] = df["Betalningsdag"]
    normalized_df["Raw Payee"] = df["Mottagare/Betalare"]
    normalized_df["Raw Bank Message"] = df["Meddelande"]
    normalized_df["Raw Amount"] = df["Belopp"]
    normalized_df["Raw Balance"] = None
    normalized_df["Real Date"] = normalized_df["Raw Real Date"].apply(
        nordea_fi_reference_number_to_datetime_obj
    )
    normalized_df["Bank Date"] = normalized_df["Raw Bank Date"].apply(
        fi_date_to_datetime_obj
    )
    normalized_df["Payee"] = normalized_df["Raw Payee"]
    normalized_df["Bank Message"] = normalized_df["Raw Bank Message"]
    normalized_df["Amount"] = normalized_df["Raw Amount"].apply(
        convert_european_amount_to_decimal
    )
    normalized_df["Balance"] = None
    normalized_df["Original data"] = df[
        [
            "Bokningsdag",
            "Valutadag",
            "Betalningsdag",
            "Belopp",
            "Mottagare/Betalare",
            "Kontonummer",
            "BIC",
            "Kontotransaktion",
            "Referens",
            "Betalarens referens",
            "Meddelande",
            "Kortets nummer",
            "Kvitto",
        ]
    ].to_dict(orient="records")
    return normalized_df[
        [
            "Real Date",
            "Bank Date",
            "Payee",
            "Bank Message",
            "Amount",
            "Balance",
            "Original data",
            "Raw Real Date",
            "Raw Bank Date",
            "Raw Payee",
            "Raw Bank Message",
            "Raw Amount",
            "Raw Balance",
        ]
    ]


def nordea_fi_lang_se_txt_transactions_parser(transaction_file):
    # type: (str) -> DataFrame
    df = import_nordea_fi_lang_se_txt_transaction_file(transaction_file)
    return nordea_fi_lang_se_txt_transactions_to_general_clerk_format(df)



import numpy as np
import pandas as pd

from clerkai.time_tracking.parse import (
    naive_time_tracking_entry_id_duplicate_nums, naive_time_tracking_entry_ids,
    time_tracking_entry_ids)

time_tracking_entries_df = pd.DataFrame(
    {
        "Raw UTC Timestamp": ["2020-04-05 09:37"],
        "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
        "UTC Timestamp": ["2020-04-05 09:37"],
        "Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Session": ["start 2020-04-05 (+0300) 09:00"],
    }
)

time_tracking_entries_with_none_df = pd.DataFrame(
    {
        "Raw UTC Timestamp": [None],
        "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
        "UTC Timestamp": [None],
        "Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Session": ["start 2020-04-05 (+0300) 09:00"],
    }
)

time_tracking_entries_with_nan_df = pd.DataFrame(
    {
        "Raw UTC Timestamp": [float("NaN")],
        "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
        "UTC Timestamp": [float("NaN")],
        "Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Session": ["start 2020-04-05 (+0300) 09:00"],
    }
)

time_tracking_entries_without_raw_columns_df = pd.DataFrame(
    {
        "UTC Timestamp": [None],
        "Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Session": ["start 2020-04-05 (+0300) 09:00"],
    }
)

time_tracking_entries_without_raw_columns_and_some_core_columns_df = pd.DataFrame(
    {
        "Source Lines Summary": ["foo | bar | 123 | zoo"],
        "Session": ["start 2020-04-05 (+0300) 09:00"],
    }
)


def test_naive_time_tracking_entry_ids():
    df = time_tracking_entries_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "Raw UTC Timestamp": ["2020-04-05 09:37"],
            "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
            "UTC Timestamp": ["2020-04-05 09:37"],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) '
                '09:00", "source_lines_summary": "F162", '
                '"utc_timestamp": "2020-04-05 09:37"}',
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_time_tracking_entry_ids_with_none():
    df = time_tracking_entries_with_none_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "Raw UTC Timestamp": [None],
            "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
            "UTC Timestamp": [None],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) 09:00", "source_lines_summary": "F162"'
                ', "utc_timestamp": null}'
            ],
        }
    )
    assert df.replace({np.nan: None}).to_dict(orient="records") == expected.replace(
        {np.nan: None}
    ).to_dict(orient="records")


def test_naive_time_tracking_entry_ids_with_nan():
    df = time_tracking_entries_with_nan_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "Raw UTC Timestamp": [float("NaN")],
            "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
            "UTC Timestamp": [float("NaN")],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) 09:00", "source_lines_summary": "F162"'
                ', "utc_timestamp": null}'
            ],
        }
    )
    assert df.replace({np.nan: None}).to_dict(orient="records") == expected.replace(
        {np.nan: None}
    ).to_dict(orient="records")


def test_naive_time_tracking_entry_ids_without_raw_columns():
    df = time_tracking_entries_without_raw_columns_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "UTC Timestamp": [None],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) 09:00", "source_lines_summary": "F162"'
                ', "utc_timestamp": null}'
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_time_tracking_entry_ids_without_raw_columns_and_some_core_columns():
    df = time_tracking_entries_without_raw_columns_and_some_core_columns_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) 09:00", "source_lines_summary": "F162"'
                ', "utc_timestamp": null}'
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_add_duplicate_num():
    import pandas as pd

    df = pd.DataFrame(
        {
            "foo": [0, 0, 0, 3, 3, 5, 5],
            "f_key": [1, 2, 1, 2, 3, 1, 1],
            "values": ["red", "blue", "green", "yellow", "orange", "violet", "cyan"],
        }
    )

    df["duplicate_num"] = df.groupby(["foo", "f_key"]).cumcount() + 1

    expected = pd.DataFrame(
        {
            "foo": [0, 0, 0, 3, 3, 5, 5],
            "f_key": [1, 2, 1, 2, 3, 1, 1],
            "values": ["red", "blue", "green", "yellow", "orange", "violet", "cyan"],
            "duplicate_num": [1, 1, 2, 1, 1, 1, 2],
        }
    )

    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_time_tracking_entry_id_duplicate_nums():
    df = time_tracking_entries_df.copy()
    df["naive_time_tracking_entry_id"] = naive_time_tracking_entry_ids(df)
    df2 = df.copy()
    df2[
        "naive_time_tracking_entry_id_duplicate_num"
    ] = naive_time_tracking_entry_id_duplicate_nums(df2)
    expected = pd.DataFrame(
        {
            "Raw UTC Timestamp": ["2020-04-05 09:37"],
            "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
            "UTC Timestamp": ["2020-04-05 09:37"],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "naive_time_tracking_entry_id": [
                '{"session": "start 2020-04-05 (+0300) '
                '09:00", "source_lines_summary": "F162", '
                '"utc_timestamp": "2020-04-05 09:37"}',
            ],
            "naive_time_tracking_entry_id_duplicate_num": [1],
        }
    )
    assert df2.to_dict(orient="records") == expected.to_dict(orient="records")


def test_time_tracking_entry_ids():
    df = time_tracking_entries_df.copy()
    df["ID"] = time_tracking_entry_ids(df)
    expected = pd.DataFrame(
        {
            "Raw UTC Timestamp": ["2020-04-05 09:37"],
            "Raw Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Raw Session": ["start 2020-04-05 (+0300) 09:00"],
            "UTC Timestamp": ["2020-04-05 09:37"],
            "Source Lines Summary": ["foo | bar | 123 | zoo"],
            "Session": ["start 2020-04-05 (+0300) 09:00"],
            "ID": [
                '{"ref": {"session": "start 2020-04-05 (+0300) 09:00", "source_lines_summary": "F162"'
                ', "utc_timestamp": "2020-04-05 09:37"}, "ord": 1}'
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")



from os.path import dirname, join, realpath

from clerkai.transactions.parsers.ee.lhv.csv import (
    lhv_ee_csv_transactions_parser, lhv_ee_description_to_datetime_obj)

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_lhv_ee_description_to_datetime_obj():
    assert (
        lhv_ee_description_to_datetime_obj(
            "(..1234) 2019-04-06 19:15 ALEPA LENTOASEMA \\LENTOASEMA T2 2B SAAP \\VANTAA \\01530 FINFIN"
        ).strftime("%Y-%m-%d %H:%M")
        == "2019-04-06 19:15"
    )


def test_lhv_ee_csv_transactions_parser():
    # type: () -> None
    transaction_file_path = join(
        test_data_dir_path,
        "EE123456789012345678_Account_Statement_2019-08-17.edited.csv",
    )
    transactions_df = lhv_ee_csv_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



import json

import pandas as pd

from clerkai.transactions.parsers.ee.lhv.csv import \
    lhv_ee_csv_transactions_parser
from clerkai.transactions.parsers.fi.nordea.personal.txt import \
    nordea_fi_lang_se_txt_transactions_parser
from clerkai.transactions.parsers.international.n26.csv import \
    n26_csv_transactions_parser
from clerkai.transactions.parsers.international.nordea.netbank.csv import \
    nordea_netbank_csv_transactions_parser
from clerkai.transactions.parsers.international.xolo.csv import \
    xolo_csv_transactions_parser
from clerkai.transactions.parsers.se.banknorwegian.xlsx import \
    banknorwegian_se_xlsx_transactions_parser
from clerkai.transactions.parsers.se.danskebank.personal.csv import \
    danskebank_se_csv_transactions_parser
from clerkai.transactions.parsers.se.nordea.personal.internetbanken_privat.xlsx import \
    nordea_se_personal_internetbanken_privat_xlsx_transactions_parser
from clerkai.utils import clerkai_input_file_path, is_nan, raw_if_available

parser_by_content_type = {
    "exported-transaction-file/nordea.fi.natbanken-privat.xls": nordea_fi_lang_se_txt_transactions_parser,
    "exported-transaction-file/lhv.ee.account-statement.csv": lhv_ee_csv_transactions_parser,
    "exported-transaction-file/nordea.se.internetbanken-privat.xls": (
        nordea_se_personal_internetbanken_privat_xlsx_transactions_parser
    ),
    "exported-transaction-file/nordea.netbank.csv": nordea_netbank_csv_transactions_parser,
    "exported-transaction-file/danskebank.se.csv": danskebank_se_csv_transactions_parser,
    "exported-transaction-file/xolo.io.expenses.csv": xolo_csv_transactions_parser,
    "exported-transaction-file/n26.com.csv": n26_csv_transactions_parser,
    "exported-transaction-file/avanza.se.transaktioner.csv": None,
    "exported-transaction-file/norwegianreward.se.via-dataminer.xlsx": None,
    "exported-transaction-file/nordea.se.internetbanken-foretag.xls": None,
    "exported-transaction-file/banknorwegian.se.xlsx": banknorwegian_se_xlsx_transactions_parser,
    "exported-transaction-file/clerk.ai.general-transactions-format.xlsx": None,
    "exported-transaction-file/paypal.com.activity-report.csv": None,
    "exported-transaction-file/skatteverket.se.skattekonto": None,
}


def naive_transaction_ids(transactions):
    import jellyfish

    def generate_naive_transaction_id(transaction):
        def none_if_nan(x):
            if is_nan(x):
                return None
            else:
                return x

        id_key_dict = {}
        id_key_dict["real_date"] = none_if_nan(
            raw_if_available("Real Date", transaction)
        )
        id_key_dict["bank_date"] = none_if_nan(
            raw_if_available("Bank Date", transaction)
        )
        payee = none_if_nan(raw_if_available("Payee", transaction))
        bank_message = none_if_nan(raw_if_available("Bank Message", transaction))
        id_key_dict["amount"] = none_if_nan(raw_if_available("Amount", transaction))
        id_key_dict["balance"] = none_if_nan(raw_if_available("Balance", transaction))
        id_key_dict["payee"] = jellyfish.soundex(payee) if type(payee) is str else payee
        id_key_dict["bank_message"] = (
            jellyfish.soundex(bank_message)
            if type(bank_message) is str
            else bank_message
        )
        return json.dumps(id_key_dict, sort_keys=True, default=str, allow_nan=False)

    return transactions.apply(generate_naive_transaction_id, axis=1)


def naive_transaction_id_duplicate_nums(transactions):
    return transactions.groupby(["naive_transaction_id"]).cumcount() + 1


def transaction_ids(transactions):
    copy = transactions.copy()
    copy["naive_transaction_id"] = naive_transaction_ids(copy)
    copy["naive_transaction_id_duplicate_num"] = naive_transaction_id_duplicate_nums(
        copy
    )

    def generate_transaction_id(transaction):
        id_key_dict = {
            "ref": json.loads(transaction["naive_transaction_id"]),
            "ord": transaction["naive_transaction_id_duplicate_num"],
        }
        return json.dumps(id_key_dict)

    return copy.apply(generate_transaction_id, axis=1)


def parse_transaction_files(
    transaction_files, clerkai_input_folder_path, keepraw=False, failfast=False
):
    class ContentTypeNotSetError(Exception):
        pass

    class ParserNotAvailableError(Exception):
        pass

    def parse_transaction_file_row(transaction_file):
        transaction_file_path = clerkai_input_file_path(
            clerkai_input_folder_path, transaction_file
        )
        results = None
        error = None

        if failfast:
            print(
                "Context (showing since failfast=True) - transaction_file:",
                "%s/%s"
                % (transaction_file["File path"], transaction_file["File name"]),
            )

        def parse():
            content_type = transaction_file["Content type"]
            if not content_type:
                raise ContentTypeNotSetError("Transaction file has no content type set")
            if (
                content_type not in parser_by_content_type
                or not parser_by_content_type[content_type]
            ):
                raise ParserNotAvailableError(
                    "Content type '%s' has no parser" % content_type
                )
            parser = parser_by_content_type[content_type]
            # print(parser, transaction_file_path)
            transactions = parser(transaction_file_path)
            transactions["Source transaction file index"] = transaction_file.name
            # add future join/merge index
            transactions["ID"] = transaction_ids(transactions)
            # drop raw columns
            if not keepraw:
                transactions = transactions.drop(
                    [
                        "Raw Real Date",
                        "Raw Bank Date",
                        "Raw Payee",
                        "Raw Bank Message",
                        "Raw Amount",
                        "Raw Currency",
                        "Raw Balance",
                        "Raw Foreign Currency",
                        "Raw Foreign Currency Amount",
                        "Raw Foreign Currency Rate",
                        "Raw Doc Status",
                        "Raw Payment Status",
                    ],
                    axis=1,
                    errors="ignore",
                )
            return transactions

        # failfast raises errors except expected/benign value errors
        if failfast:
            try:
                results = parse()
            except (ContentTypeNotSetError, ParserNotAvailableError) as e:
                error = e
        else:
            try:
                results = parse()
            except Exception as e:
                error = e
        return pd.Series([results, error], index=["Parse results", "Error"])

    if len(transaction_files) == 0:
        raise Exception("No transaction files to parse")

    parsed_transaction_file_results = transaction_files.apply(
        parse_transaction_file_row, axis=1
    )

    parsed_transaction_files = transaction_files.join(parsed_transaction_file_results)
    return parsed_transaction_files


def transactions_from_parsed_transaction_files(parsed_transaction_files):
    transactions_df = pd.concat(
        parsed_transaction_files["Parse results"].values, sort=False
    ).reset_index(drop=True)
    transactions_df = pd.merge(
        transactions_df,
        parsed_transaction_files.drop(
            ["Parse results", "History reference"], axis=1
        ).add_prefix("Source transaction file: "),
        left_on="Source transaction file index",
        right_index=True,
    )
    return transactions_df.drop(["Source transaction file index"], axis=1)



import os

from clerkai.location_history.defaults import (
    location_history_by_date_editable_columns,
    location_history_files_editable_columns)
from clerkai.location_history.flow import location_history_flow
from clerkai.time_tracking.defaults import (
    default_time_tracking_entries_editable_columns,
    default_time_tracking_files_editable_columns)
from clerkai.time_tracking.flow import time_tracking_flow
from clerkai.transactions.defaults import (
    default_transaction_files_editable_columns,
    default_transactions_editable_columns)
from clerkai.transactions.flow import transactions_flow
from clerkai.utils import (add_all_untracked_and_changed_files,
                           commit_datetime_from_history_reference,
                           commits_by_short_gitsha1, current_gitsha1,
                           ensure_clerkai_folder_versioning,
                           export_file_name_by_record_type,
                           fetch_gsheets_worksheet_as_df,
                           list_files_in_clerk_input_subfolder,
                           list_files_in_clerk_subfolder,
                           possibly_edited_commit_specific_df,
                           possibly_edited_df_util)


def extract_commit_sha_from_edit_subfolder_path(edit_subfolder_path):
    import re

    p = re.compile("\\(([^)]*)\\)", re.IGNORECASE)
    m = p.search(edit_subfolder_path)
    commit_sha = None
    if len(m.groups()) > 0:
        commit_sha = m.groups()[0]
    return commit_sha


def init_notebook_and_return_helpers(clerkai_folder):
    # expand given paths to absolute paths
    clerkai_folder_path = os.path.expanduser(clerkai_folder).rstrip(os.sep)
    clerkai_input_folder_path = os.path.join(clerkai_folder_path, "Input")
    transactions_folder_path = os.path.join(clerkai_input_folder_path, "Transactions")
    receipts_folder_path = os.path.join(clerkai_input_folder_path, "Receipts")
    time_tracking_folder_path = os.path.join(clerkai_input_folder_path, "Time Tracking")
    location_history_folder_path = os.path.join(
        clerkai_input_folder_path, "Location History"
    )
    edits_folder_path = os.path.join(clerkai_folder_path, "Edits")

    # set working dir to be the clerk.ai folder
    os.chdir(clerkai_folder_path)

    # initiate / validate clerk.ai-folder versioning
    clerkai_input_folder_repo = ensure_clerkai_folder_versioning(
        clerkai_input_folder_path=clerkai_input_folder_path
    )

    def acknowledge_changes_in_clerkai_input_folder():
        add_all_untracked_and_changed_files(clerkai_input_folder_repo)

    def current_history_reference():
        return current_gitsha1(clerkai_input_folder_repo)

    # transactions

    def transactions(
        keep_unmerged_previous_edits=False,
        failfast=False,
        additional_transaction_files_editable_columns=None,
        additional_transactions_editable_columns=None,
    ):
        if additional_transaction_files_editable_columns:
            transaction_files_editable_columns = [
                *additional_transaction_files_editable_columns,
                *default_transaction_files_editable_columns,
            ]
        else:
            transaction_files_editable_columns = (
                default_transaction_files_editable_columns
            )
        if additional_transactions_editable_columns:
            transactions_editable_columns = [
                *additional_transactions_editable_columns,
                *default_transactions_editable_columns,
            ]
        else:
            transactions_editable_columns = default_transactions_editable_columns
        return transactions_flow(
            transaction_files_editable_columns=transaction_files_editable_columns,
            transactions_editable_columns=transactions_editable_columns,
            clerkai_input_folder_path=clerkai_input_folder_path,
            possibly_edited_df=possibly_edited_df,
            transactions_folder_path=transactions_folder_path,
            acknowledge_changes_in_clerkai_input_folder=acknowledge_changes_in_clerkai_input_folder,
            current_history_reference=current_history_reference,
            keep_unmerged_previous_edits=keep_unmerged_previous_edits,
            failfast=failfast,
        )

    # receipts

    def list_receipt_files_in_receipts_folder():
        _ = list_files_in_clerk_input_subfolder(
            receipts_folder_path, clerkai_input_folder_path=clerkai_input_folder_path
        )
        if len(_) == 0:
            return _
        _["Ignore"] = None
        _["History reference"] = current_history_reference()
        return _[
            ["File name", "File path", "Ignore", "File metadata", "History reference"]
        ]

    # location_history

    def location_history(keep_unmerged_previous_edits=False, failfast=False):
        return location_history_flow(
            location_history_files_editable_columns=location_history_files_editable_columns,
            location_history_by_date_editable_columns=location_history_by_date_editable_columns,
            clerkai_input_folder_path=clerkai_input_folder_path,
            possibly_edited_df=possibly_edited_df,
            location_history_folder_path=location_history_folder_path,
            acknowledge_changes_in_clerkai_input_folder=acknowledge_changes_in_clerkai_input_folder,
            current_history_reference=current_history_reference,
            keep_unmerged_previous_edits=keep_unmerged_previous_edits,
            failfast=failfast,
        )

    # time_tracking_entries

    def time_tracking_entries(
        keep_unmerged_previous_edits=False,
        failfast=False,
        additional_time_tracking_files_editable_columns=None,
        additional_time_tracking_entries_editable_columns=None,
    ):
        if additional_time_tracking_files_editable_columns:
            time_tracking_files_editable_columns = [
                *additional_time_tracking_files_editable_columns,
                *default_time_tracking_files_editable_columns,
            ]
        else:
            time_tracking_files_editable_columns = (
                default_time_tracking_files_editable_columns
            )
        if additional_time_tracking_entries_editable_columns:
            time_tracking_entries_editable_columns = [
                *additional_time_tracking_entries_editable_columns,
                *default_time_tracking_entries_editable_columns,
            ]
        else:
            time_tracking_entries_editable_columns = (
                default_time_tracking_entries_editable_columns
            )
        return time_tracking_flow(
            time_tracking_files_editable_columns=time_tracking_files_editable_columns,
            time_tracking_entries_editable_columns=time_tracking_entries_editable_columns,
            clerkai_input_folder_path=clerkai_input_folder_path,
            possibly_edited_df=possibly_edited_df,
            time_tracking_folder_path=time_tracking_folder_path,
            acknowledge_changes_in_clerkai_input_folder=acknowledge_changes_in_clerkai_input_folder,
            current_history_reference=current_history_reference,
            keep_unmerged_previous_edits=keep_unmerged_previous_edits,
            failfast=failfast,
        )

    # other

    def list_edit_files_in_edits_folder():
        _ = list_files_in_clerk_subfolder(
            edits_folder_path, clerkai_folder_path=clerkai_folder_path
        )
        if len(_) == 0:
            return _
        # ignore those in the archive subfolder
        _ = _[~_["File path"].str.contains("\\/Archive\\/", regex=True)]
        if len(_) == 0:
            return _
        # add commit-metadata to list
        commits = commits_by_short_gitsha1(
            clerkai_input_folder_path, clerkai_input_folder_repo
        )
        _["Related history reference"] = _["File path"].apply(
            extract_commit_sha_from_edit_subfolder_path
        )
        _["Related history reference date"] = _["Related history reference"].apply(
            commit_datetime_from_history_reference, commits=commits
        )
        _ = _.sort_values(by="Related history reference date")
        return _

    # ensure_directories_are_in_place()
    if not os.path.isdir(transactions_folder_path):
        os.mkdir(transactions_folder_path)
    # if not os.path.isdir(receipts_folder_path):
    #    os.mkdir(receipts_folder_path)
    if not os.path.isdir(location_history_folder_path):
        os.mkdir(location_history_folder_path)
    if not os.path.isdir(time_tracking_folder_path):
        os.mkdir(time_tracking_folder_path)
    if not os.path.isdir(edits_folder_path):
        os.mkdir(edits_folder_path)

    def possibly_edited_df(
        current_commit_df,
        record_type,
        editable_columns,
        keep_unmerged_previous_edits=False,
    ):
        return possibly_edited_df_util(
            current_commit_df,
            record_type,
            editable_columns,
            keep_unmerged_previous_edits,
            list_edit_files_in_edits_folder,
            current_history_reference,
            edits_folder_path,
            clerkai_input_folder_path,
            clerkai_input_folder_repo,
        )

    def store_gsheets_edits(gsheets_title, gsheets_sheet_name, edits_df, record_type):
        from datetime import datetime

        """
        print(
            gsheets_title,
            gsheets_sheet_name,
            edits_df.columns,
            edits_df["History reference"].unique(),
        )
        """

        history_references = edits_df["History reference"].unique()

        if len(history_references) > 1:
            raise Exception("Edited data should only contain rows from a single export")

        history_reference = str(history_references[0])

        (dt, micro) = datetime.utcnow().strftime("%Y-%m-%d %H%M%S.%f").split(".")
        timestamp = "%s%03d" % (dt, int(micro) / 1000)

        suffix = ".gsheets.%s.%s.%s" % (gsheets_title, gsheets_sheet_name, timestamp)
        (export_file_name, export_file_name_base) = export_file_name_by_record_type(
            record_type, suffix=suffix
        )

        commits = commits_by_short_gitsha1(
            clerkai_input_folder_path, clerkai_input_folder_repo
        )
        commit_datetime = commit_datetime_from_history_reference(
            history_reference, commits=commits
        )

        edit_folder_stored_edits_df = possibly_edited_commit_specific_df(
            df=edits_df,
            record_type=record_type,
            export_file_name=export_file_name,
            edits_folder_path=edits_folder_path,
            commit_datetime=commit_datetime,
            history_reference=history_reference,
            create_if_not_exists=True,
            create_if_exists=True,
        )

        return edit_folder_stored_edits_df

    def download_and_store_gsheets_edits(
        gsheets_client, gsheets_title, gsheets_sheet_name, record_type
    ):
        edits_df = fetch_gsheets_worksheet_as_df(
            gsheets_client, gsheets_title, gsheets_sheet_name
        )
        store_gsheets_edits(gsheets_title, gsheets_sheet_name, edits_df, record_type)

    return {
        "transactions": transactions,
        "list_receipt_files_in_receipts_folder": list_receipt_files_in_receipts_folder,
        "location_history": location_history,
        "time_tracking_entries": time_tracking_entries,
        "acknowledge_changes_in_clerkai_input_folder": acknowledge_changes_in_clerkai_input_folder,
        "store_gsheets_edits": store_gsheets_edits,
        "download_and_store_gsheets_edits": download_and_store_gsheets_edits,
    }



from datetime import datetime

import numpy as np
import pandas as pd

from clerkai.transactions.parse import (naive_transaction_id_duplicate_nums,
                                        naive_transaction_ids, transaction_ids)

transactions_df = pd.DataFrame(
    {
        "Raw Real Date": [None],
        "Raw Bank Date": ["2019/05/02"],
        "Raw Payee": ["Acme-Industries 123 Inc"],
        "Raw Bank Message": ["Foo"],
        "Raw Amount": ["3.000,12"],
        "Raw Balance": [None],
        "Real Date": [None],
        "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
        "Payee": ["Acme-Industries 123 Inc"],
        "Bank Message": ["Foo"],
        "Amount": [3000.12],
        "Balance": [None],
    }
)

transactions_with_nan_df = pd.DataFrame(
    {
        "Raw Real Date": [float("NaN")],
        "Raw Bank Date": ["2019/05/02"],
        "Raw Payee": ["Acme-Industries 123 Inc"],
        "Raw Bank Message": ["Foo"],
        "Raw Amount": ["3.000,12"],
        "Raw Balance": [None],
        "Real Date": [None],
        "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
        "Payee": ["Acme-Industries 123 Inc"],
        "Bank Message": ["Foo"],
        "Amount": [3000.12],
        "Balance": [None],
    }
)

transactions_without_raw_columns_df = pd.DataFrame(
    {
        "Real Date": [None],
        "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
        "Payee": ["Acme-Industries 123 Inc"],
        "Bank Message": ["Foo"],
        "Amount": [3000.12],
        "Balance": [None],
    }
)

transactions_without_raw_columns_and_some_core_columns_df = pd.DataFrame(
    {
        "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
        "Payee": ["Acme-Industries 123 Inc"],
        "Amount": [3000.12],
        "Balance": [None],
    }
)


def test_naive_transaction_ids():
    df = transactions_df.copy()
    df["naive_transaction_id"] = naive_transaction_ids(df)
    expected = pd.DataFrame(
        {
            "Raw Real Date": [None],
            "Raw Bank Date": ["2019/05/02"],
            "Raw Payee": ["Acme-Industries 123 Inc"],
            "Raw Bank Message": ["Foo"],
            "Raw Amount": ["3.000,12"],
            "Raw Balance": [None],
            "Real Date": [None],
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Bank Message": ["Foo"],
            "Amount": [3000.12],
            "Balance": [None],
            "naive_transaction_id": [
                (
                    '{"amount": "3.000,12", "balance": null, "bank_date": "2019/05/02", '
                    '"bank_message": "F000", "payee": "A255", "real_date": null}'
                )
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_transaction_ids_with_nan():
    df = transactions_with_nan_df.copy()
    df["naive_transaction_id"] = naive_transaction_ids(df)
    expected = pd.DataFrame(
        {
            "Raw Real Date": [float("NaN")],
            "Raw Bank Date": ["2019/05/02"],
            "Raw Payee": ["Acme-Industries 123 Inc"],
            "Raw Bank Message": ["Foo"],
            "Raw Amount": ["3.000,12"],
            "Raw Balance": [None],
            "Real Date": [None],
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Bank Message": ["Foo"],
            "Amount": [3000.12],
            "Balance": [None],
            "naive_transaction_id": [
                (
                    '{"amount": "3.000,12", "balance": null, "bank_date": "2019/05/02", '
                    '"bank_message": "F000", "payee": "A255", "real_date": null}'
                )
            ],
        }
    )
    assert df.replace({np.nan: None}).to_dict(orient="records") == expected.replace(
        {np.nan: None}
    ).to_dict(orient="records")


def test_naive_transaction_ids_without_raw_columns():
    df = transactions_without_raw_columns_df.copy()
    df["naive_transaction_id"] = naive_transaction_ids(df)
    expected = pd.DataFrame(
        {
            "Real Date": [None],
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Bank Message": ["Foo"],
            "Amount": [3000.12],
            "Balance": [None],
            "naive_transaction_id": [
                (
                    '{"amount": 3000.12, "balance": null, "bank_date": "2019-05-02 00:00:00", '
                    '"bank_message": "F000", "payee": "A255", "real_date": null}'
                )
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_transaction_ids_without_raw_columns_and_some_core_columns():
    df = transactions_without_raw_columns_and_some_core_columns_df.copy()
    df["naive_transaction_id"] = naive_transaction_ids(df)
    expected = pd.DataFrame(
        {
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Amount": [3000.12],
            "Balance": [None],
            "naive_transaction_id": [
                (
                    '{"amount": 3000.12, "balance": null, "bank_date": "2019-05-02 00:00:00",'
                    ' "bank_message": null, "payee": "A255", "real_date": null}'
                )
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_add_duplicate_num():
    import pandas as pd

    df = pd.DataFrame(
        {
            "foo": [0, 0, 0, 3, 3, 5, 5],
            "f_key": [1, 2, 1, 2, 3, 1, 1],
            "values": ["red", "blue", "green", "yellow", "orange", "violet", "cyan"],
        }
    )

    df["duplicate_num"] = df.groupby(["foo", "f_key"]).cumcount() + 1

    expected = pd.DataFrame(
        {
            "foo": [0, 0, 0, 3, 3, 5, 5],
            "f_key": [1, 2, 1, 2, 3, 1, 1],
            "values": ["red", "blue", "green", "yellow", "orange", "violet", "cyan"],
            "duplicate_num": [1, 1, 2, 1, 1, 1, 2],
        }
    )

    assert df.to_dict(orient="records") == expected.to_dict(orient="records")


def test_naive_transaction_id_duplicate_nums():
    df = transactions_df.copy()
    df["naive_transaction_id"] = naive_transaction_ids(df)
    df2 = df.copy()
    df2["naive_transaction_id_duplicate_num"] = naive_transaction_id_duplicate_nums(df2)
    expected = pd.DataFrame(
        {
            "Raw Real Date": [None],
            "Raw Bank Date": ["2019/05/02"],
            "Raw Payee": ["Acme-Industries 123 Inc"],
            "Raw Bank Message": ["Foo"],
            "Raw Amount": ["3.000,12"],
            "Raw Balance": [None],
            "Real Date": [None],
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Bank Message": ["Foo"],
            "Amount": [3000.12],
            "Balance": [None],
            "naive_transaction_id": [
                (
                    '{"amount": "3.000,12", "balance": null, "bank_date": "2019/05/02", '
                    '"bank_message": "F000", "payee": "A255", "real_date": null}'
                )
            ],
            "naive_transaction_id_duplicate_num": [1],
        }
    )
    assert df2.to_dict(orient="records") == expected.to_dict(orient="records")


def test_transaction_ids():
    df = transactions_df.copy()
    df["ID"] = transaction_ids(df)
    expected = pd.DataFrame(
        {
            "Raw Real Date": [None],
            "Raw Bank Date": ["2019/05/02"],
            "Raw Payee": ["Acme-Industries 123 Inc"],
            "Raw Bank Message": ["Foo"],
            "Raw Amount": ["3.000,12"],
            "Raw Balance": [None],
            "Real Date": [None],
            "Bank Date": [datetime.strptime("2019/05/02", "%Y/%m/%d")],
            "Payee": ["Acme-Industries 123 Inc"],
            "Bank Message": ["Foo"],
            "Amount": [3000.12],
            "Balance": [None],
            "ID": [
                (
                    '{"ref": {"amount": "3.000,12", "balance": null, "bank_date": "2019/05/02",'
                    ' "bank_message": "F000", "payee": "A255", "real_date": null}, "ord": 1}'
                )
            ],
        }
    )
    assert df.to_dict(orient="records") == expected.to_dict(orient="records")



from os.path import dirname, join, realpath

from clerkai.location_history.defaults import (
    location_history_by_date_editable_columns,
    location_history_files_editable_columns)
from clerkai.location_history.flow import location_history_flow

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_location_history_flow():
    def acknowledge_changes_in_clerkai_input_folder():
        pass

    def current_history_reference():
        return "foo"

    def possibly_edited_df(
        current_commit_df,
        record_type,
        editable_columns,
        keep_unmerged_previous_edits=False,
    ):
        if record_type == "location_history_files":
            current_commit_df.loc[
                current_commit_df["File name"] == "Location History.edited.json",
                "Content type",
            ] = "exported-location-history-file/google-takeout.location-history.json"
            current_commit_df.loc[
                current_commit_df["File name"] == "dropbox-camera-uploads.edited.csv",
                "Content type",
            ] = "exported-location-history-file/exiftool-output.csv"
        return current_commit_df

    clerkai_input_folder_path = join(test_data_dir_path, "Input")
    location_history_folder_path = join(clerkai_input_folder_path, "Location History")

    (
        location_history_files_df,
        possibly_edited_location_history_files_df,
        unsuccessfully_parsed_location_history_files,
        successfully_parsed_location_history_files,
        all_parsed_location_history_df,
        location_history_df,
        location_history_with_geonames_df,
        possibly_edited_location_history_by_date_df,
    ) = location_history_flow(
        location_history_files_editable_columns=location_history_files_editable_columns,
        location_history_by_date_editable_columns=location_history_by_date_editable_columns,
        clerkai_input_folder_path=clerkai_input_folder_path,
        possibly_edited_df=possibly_edited_df,
        location_history_folder_path=location_history_folder_path,
        acknowledge_changes_in_clerkai_input_folder=acknowledge_changes_in_clerkai_input_folder,
        current_history_reference=current_history_reference,
        keep_unmerged_previous_edits=False,
        failfast=True,
    )

    assert len(unsuccessfully_parsed_location_history_files) == 0
    assert len(possibly_edited_location_history_by_date_df) > 0
    actual = possibly_edited_location_history_by_date_df.to_csv(index=False)
    possibly_edited_location_history_by_date_df_csv_path = join(
        test_data_dir_path, "possibly_edited_location_history_by_date_df"
    )
    actual_file_path = "%s%s" % (
        possibly_edited_location_history_by_date_df_csv_path,
        ".actual.csv",
    )
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (
        possibly_edited_location_history_by_date_df_csv_path,
        ".expected.csv",
    )
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



import pandas as pd

from clerkai.location_history.parsers.exiftool.csv import \
    exiftool_csv_location_history_parser
from clerkai.location_history.parsers.google.takeout.json import \
    google_takeout_location_history_json_location_history_parser
from clerkai.utils import clerkai_input_file_path

parser_by_content_type = {
    "exported-location-history-file/exiftool-output.csv": exiftool_csv_location_history_parser,
    (
        "exported-location-history-file/google-takeout.location-history.json"
    ): google_takeout_location_history_json_location_history_parser,
}


def parse_location_history_files(
    location_history_files, clerkai_input_folder_path, keepraw=False, failfast=False
):
    class ContentTypeNotSetError(Exception):
        pass

    class ParserNotAvailableError(Exception):
        pass

    def parse_location_history_file_row(location_history_file):
        location_history_file_path = clerkai_input_file_path(
            clerkai_input_folder_path, location_history_file
        )
        results = None
        error = None

        if failfast:
            print(
                "location_history_file",
                "%s/%s"
                % (
                    location_history_file["File path"],
                    location_history_file["File name"],
                ),
            )

        def parse():
            content_type = location_history_file["Content type"]
            if not content_type:
                raise ContentTypeNotSetError("Transaction file has no content type set")
            if (
                content_type not in parser_by_content_type
                or not parser_by_content_type[content_type]
            ):
                raise ParserNotAvailableError(
                    "Content type '%s' has no parser" % content_type
                )
            parser = parser_by_content_type[content_type]
            # print(parser, location_history_file_path)
            location_history = parser(location_history_file_path)
            location_history[
                "Source location history file index"
            ] = location_history_file.name
            # drop raw columns
            if not keepraw:
                location_history = location_history.drop(
                    [
                        "Raw Real Date",
                        "Raw Bank Date",
                        "Raw Payee",
                        "Raw Bank Message",
                        "Raw Amount",
                        "Raw Currency",
                        "Raw Balance",
                        "Raw Foreign Currency",
                        "Raw Foreign Currency Amount",
                        "Raw Foreign Currency Rate",
                        "Raw Doc Status",
                        "Raw Payment Status",
                    ],
                    axis=1,
                    errors="ignore",
                )
            return location_history

        # failfast raises errors except expected/benign value errors
        if failfast:
            try:
                results = parse()
            except (ContentTypeNotSetError, ParserNotAvailableError) as e:
                error = e
        else:
            try:
                results = parse()
            except Exception as e:
                error = e
        return pd.Series([results, error], index=["Parse results", "Error"])

    if len(location_history_files) == 0:
        raise Exception("No transaction files to parse")

    parsed_location_history_file_results = location_history_files.apply(
        parse_location_history_file_row, axis=1
    )

    parsed_location_history_files = location_history_files.join(
        parsed_location_history_file_results
    )
    return parsed_location_history_files


def location_history_from_parsed_location_history_files(parsed_location_history_files):
    location_history_df = pd.concat(
        parsed_location_history_files["Parse results"].values, sort=False
    ).reset_index(drop=True)
    location_history_df = pd.merge(
        location_history_df,
        parsed_location_history_files.drop(
            ["Parse results", "History reference"], axis=1
        ).add_prefix("Source transaction file: "),
        left_on="Source transaction file index",
        right_index=True,
    )
    return location_history_df.drop(["Source transaction file index"], axis=1)



import pandas as pd
from pandas.core.frame import DataFrame

from clerkai.transactions.parsers.parse_utils import amount_to_rounded_decimal
from clerkai.utils import ymd_date_to_naive_datetime_obj


def import_n26_csv_transaction_file(transaction_file):
    # type: (str) -> DataFrame
    return pd.read_csv(transaction_file)


def n26_csv_transactions_to_general_clerk_format(df):
    # type: (DataFrame) -> DataFrame
    normalized_df = pd.DataFrame()
    normalized_df["Raw Real Date"] = None
    normalized_df["Raw Bank Date"] = df["Date"]
    normalized_df["Raw Payee"] = df["Payee"]
    normalized_df["Raw Bank Message"] = df["Payment reference"]
    normalized_df["Raw Amount"] = df["Amount (EUR)"]
    normalized_df["Raw Currency"] = None
    normalized_df["Raw Balance"] = None
    normalized_df["Raw Foreign Currency"] = df["Type Foreign Currency"]
    normalized_df["Raw Foreign Currency Amount"] = df["Amount (Foreign Currency)"]
    normalized_df["Raw Foreign Currency Rate"] = df["Exchange Rate"]
    normalized_df["Real Date"] = normalized_df["Raw Real Date"]
    normalized_df["Bank Date"] = normalized_df["Raw Bank Date"].apply(
        ymd_date_to_naive_datetime_obj
    )
    normalized_df["Payee"] = normalized_df["Raw Payee"]
    normalized_df["Bank Message"] = normalized_df["Raw Bank Message"]
    normalized_df["Amount"] = normalized_df["Raw Amount"].apply(
        amount_to_rounded_decimal
    )
    normalized_df["Currency"] = "EUR"
    normalized_df["Balance"] = None
    normalized_df["Foreign Currency"] = normalized_df["Raw Foreign Currency"]
    normalized_df["Foreign Currency Amount"] = normalized_df[
        "Raw Foreign Currency Amount"
    ].apply(amount_to_rounded_decimal)
    normalized_df["Foreign Currency Rate"] = normalized_df[
        "Raw Foreign Currency Rate"
    ].apply(amount_to_rounded_decimal, accuracy=14)
    normalized_df["Original data"] = df[
        [
            "Date",
            "Payee",
            "Account number",
            "Transaction type",
            "Payment reference",
            "Category",
            "Amount (EUR)",
            "Amount (Foreign Currency)",
            "Type Foreign Currency",
            "Exchange Rate",
        ]
    ].to_dict(orient="records")
    return normalized_df[
        [
            "Real Date",
            "Bank Date",
            "Payee",
            "Bank Message",
            "Amount",
            "Currency",
            "Balance",
            "Foreign Currency",
            "Foreign Currency Amount",
            "Foreign Currency Rate",
            "Original data",
            "Raw Real Date",
            "Raw Bank Date",
            "Raw Payee",
            "Raw Bank Message",
            "Raw Amount",
            "Raw Currency",
            "Raw Balance",
            "Raw Foreign Currency",
            "Raw Foreign Currency Amount",
            "Raw Foreign Currency Rate",
        ]
    ]


def n26_csv_transactions_parser(transaction_file):
    # type: (str) -> DataFrame
    df = import_n26_csv_transaction_file(transaction_file)
    return n26_csv_transactions_to_general_clerk_format(df)



from os.path import dirname, join, realpath

from clerkai.transactions.parsers.international.nordea.netbank.csv import \
    nordea_netbank_csv_transactions_parser

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_nordea_netbank_csv_transactions_parser_se_lang_sv():
    # type: () -> None
    transaction_file_path = join(
        test_data_dir_path, "PERSONKONTO 1234 56 78901 - 2020.01.25 16.47.edited.csv"
    )
    transactions_df = nordea_netbank_csv_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected


def test_nordea_netbank_csv_transactions_parser_fi_lang_sv():
    # type: () -> None
    transaction_file_path = join(
        test_data_dir_path,
        "Personkonto FI12 3456 7890 1234 56 - 2020.05.30 17.51 - sv.edited.csv",
    )
    transactions_df = nordea_netbank_csv_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected


def test_nordea_netbank_csv_transactions_parser_fi_lang_fi():
    # type: () -> None
    transaction_file_path = join(
        test_data_dir_path,
        "Personkonto FI12 3456 7890 1234 56 - 2020.05.30 17.47 - fi.edited.csv",
    )
    transactions_df = nordea_netbank_csv_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected


def test_nordea_netbank_csv_transactions_parser_fi_lang_en():
    # type: () -> None
    transaction_file_path = join(
        test_data_dir_path,
        "Personkonto FI12 3456 7890 1234 56 - 2020.05.30 17.49 - en.edited.csv",
    )
    transactions_df = nordea_netbank_csv_transactions_parser(transaction_file_path)
    assert not transactions_df.empty
    actual = transactions_df.to_csv(index=False)
    actual_file_path = "%s%s" % (transaction_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (transaction_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



from os.path import dirname, join, realpath

from clerkai.location_history.parsers.google.takeout.json import \
    google_takeout_location_history_json_location_history_parser

test_data_dir_path = join(dirname(realpath(__file__)), "test_data")


def test_google_takeout_location_history_json_location_history_parser():
    # type: () -> None
    location_history_file_path = join(
        test_data_dir_path, "Location History.edited.json"
    )
    location_history_df = google_takeout_location_history_json_location_history_parser(
        location_history_file_path
    )
    assert not location_history_df.empty
    actual = location_history_df.to_csv(index=False)
    actual_file_path = "%s%s" % (location_history_file_path, ".actual.csv")
    with open(actual_file_path, "w") as f:
        f.write(actual)
    expected_file_path = "%s%s" % (location_history_file_path, ".expected.csv")
    with open(expected_file_path, "r") as f:
        expected = f.read()
    assert actual == expected



import json
import subprocess
from datetime import datetime
from typing import List

import pandas as pd
import pytz
from pandas import json_normalize
from pandas.core.frame import DataFrame

from clerkai.utils import is_nan, ymd_date_to_naive_datetime_obj

return_columns: List[str] = [
    "Raw Ignore",
    "Raw Source Line Numbers",
    "Raw Source Lines Summary",
    "Raw Session",
    "Raw UTC Timestamp",
    "Raw Work Date",
    # "Raw Timestamp Before Parsing",
    "Raw Timezone",
    "Raw Time-annotated Source Lines Summary",
    "Raw Hours",
    # "Raw Hours Rounded",
    "Raw Annotation",
    "Raw Year",
    "Raw Year-half",
    "Raw Month",
    "Raw Week",
    "Raw Client",
    "Raw Invoice",
    "Raw Invoice row ordinal",
    "Raw Invoice row item",
    "Raw Invoice row item category",
    "Raw Invoice row item reference",
    "Raw Invoice hours",
    "Ignore",
    "Source Line Numbers",
    "Source Lines Summary",
    "Session",
    "UTC Timestamp",
    "Work Date",
    # "Timestamp Before Parsing",
    "Timezone",
    "Time-annotated Source Lines Summary",
    "Hours",
    # "Hours Rounded",
    "Annotation",
    "Year",
    "Year-half",
    "Month",
    "Week",
    "Client",
    "Invoice",
    "Invoice row ordinal",
    "Invoice row item",
    "Invoice row item category",
    "Invoice row item reference",
    "Invoice hours",
    "Original data",
]


def neamtime_datetime_to_utc_datetime_obj(neamtime_datetime):
    naive_parsed_datetime = neamtime_datetime_to_naive_datetime_obj(neamtime_datetime)
    parsed_datetime = pytz.utc.localize(naive_parsed_datetime)
    return parsed_datetime


def neamtime_datetime_to_naive_datetime_obj(neamtime_datetime):
    if neamtime_datetime is None:
        return None
    if is_nan(neamtime_datetime):
        return None
    naive_parsed_datetime = datetime.strptime(neamtime_datetime, "%Y-%m-%d %H:%M")
    return naive_parsed_datetime


def parse_neamtime_tslog_time_tracking_file(time_tracking_file_path):
    # type: (str) -> DataFrame
    # print("time_tracking_file_path", time_tracking_file_path)

    result = subprocess.run(
        ["neamtime-log-parser", "--filePath", time_tracking_file_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if result.returncode == 1:
        print("Neamtime log parsing failed: %s" % result.stderr.decode("utf-8"))
        raise Exception("Neamtime log parsing failed")

    # print("result", result)

    json_str = result.stdout.decode("utf-8")
    # print("json_str", json_str)

    parse_results = json.loads(json_str)
    # print("parse_results", parse_results)

    parsed_time_log_entries = json_normalize(
        parse_results["timeLogEntriesWithMetadata"]
    )
    del parse_results["timeLogEntriesWithMetadata"]

    # summarize processing errors - only pick out the most relevant ones
    processing_errors = pd.DataFrame()
    processing_errors_with_fluff = json_normalize(parse_results["processingErrors"])
    del parse_results["processingErrors"]
    if (
        "thrownException" in parse_results
        and "processedTimeSpendingLog" in parse_results["thrownException"]
    ):
        del parse_results["thrownException"]["processedTimeSpendingLog"]

    if "ref" in processing_errors_with_fluff:
        issues_during_initial_parsing_df = processing_errors_with_fluff[
            processing_errors_with_fluff["ref"] == "issues-during-initial-parsing"
        ]
        if len(issues_during_initial_parsing_df) > 0:
            issues_during_initial_parsing = issues_during_initial_parsing_df.iloc[0]
            # print("issues_during_initial_parsing.columns", issues_during_initial_parsing.columns)
            issues_during_initial_parsing_data = issues_during_initial_parsing["data"]
            # print("issues_during_initial_parsing_data", issues_during_initial_parsing_data)
            processing_errors = json_normalize(issues_during_initial_parsing_data)

    # print("tslog.py - processing_errors_with_fluff.columns", processing_errors_with_fluff.columns)
    # print("tslog.py - processing_errors", processing_errors)
    # print("tslog.py - processing_errors.columns", processing_errors.columns)

    # TODO: use time_report_data? currently ignoring
    # time_report_data = parse_results["timeReportData"]
    # print("time_report_data", time_report_data)
    del parse_results["timeReportData"]
    parsing_metadata = json_normalize(parse_results)

    # print("tslog.py - parsing_metadata", parsing_metadata)
    # print("tslog.py - parsing_metadata.columns", parsing_metadata.columns)

    # parsing_metadata["timeReportData"] = None
    # parsing_metadata["timeReportData"].iloc[0] = time_report_data

    # print("tslog.py - parsed_time_log_entries", parsed_time_log_entries)
    # print("tslog.py - parsed_time_log_entries.columns", parsed_time_log_entries.columns)

    return parsed_time_log_entries, parsing_metadata, processing_errors


def neamtime_tslog_time_tracking_entries_to_general_clerk_format(df):
    # type: (DataFrame) -> DataFrame
    normalized_df = pd.DataFrame(columns=return_columns)

    if len(df) == 0:
        return normalized_df

    """
    normalized_df["Raw "] = df["ts"]
    normalized_df["Raw "] = df["sessionMeta.tzFirst"]
    """

    normalized_df["Raw Ignore"] = None
    normalized_df["Raw Source Line Numbers"] = None
    normalized_df["Raw Source Lines Summary"] = df["text"]
    normalized_df["Raw Session"] = df["sessionMeta.session_ref"]
    normalized_df["Raw UTC Timestamp"] = df["gmtTimestamp"]
    normalized_df["Raw Work Date"] = df["date"]
    # normalized_df["Raw Timestamp Before Parsing"] = df["dateRaw"] # No, dateRaw is not what it seems
    normalized_df["Raw Timezone"] = df["tz"]
    normalized_df["Raw Time-annotated Source Lines Summary"] = df["lineWithoutDate"]
    normalized_df["Raw Hours"] = df["hours"]
    # normalized_df["Raw Hours Rounded"] = df["hoursRounded"]
    normalized_df["Raw Annotation"] = df["category"]
    normalized_df["Raw Year"] = None
    normalized_df["Raw Year-half"] = None
    normalized_df["Raw Month"] = None
    normalized_df["Raw Week"] = None
    normalized_df["Raw Client"] = None
    normalized_df["Raw Invoice"] = None
    normalized_df["Raw Invoice row ordinal"] = None
    normalized_df["Raw Invoice row item"] = None
    normalized_df["Raw Invoice row item category"] = None
    normalized_df["Raw Invoice row item reference"] = None
    normalized_df["Raw Invoice hours"] = None

    normalized_df["Ignore"] = normalized_df["Raw Ignore"]
    normalized_df["Source Line Numbers"] = normalized_df["Raw Source Line Numbers"]
    normalized_df["Source Lines Summary"] = normalized_df[
        "Raw Source Lines Summary"
    ].str.strip()
    normalized_df["Session"] = normalized_df["Raw Session"]
    normalized_df["UTC Timestamp"] = normalized_df["Raw UTC Timestamp"].apply(
        neamtime_datetime_to_naive_datetime_obj
    )
    normalized_df["Work Date"] = normalized_df["Raw Work Date"].apply(
        ymd_date_to_naive_datetime_obj
    )
    # normalized_df["Timestamp Before Parsing"] = normalized_df["Raw Timestamp Before Parsing"]
    normalized_df["Timezone"] = normalized_df["Raw Timezone"]
    normalized_df["Time-annotated Source Lines Summary"] = normalized_df[
        "Raw Time-annotated Source Lines Summary"
    ].str.strip()
    normalized_df["Hours"] = normalized_df["Raw Hours"]
    # normalized_df["Hours Rounded"] = normalized_df["Raw Hours Rounded"].apply(
    #     amount_to_rounded_decimal
    # )
    normalized_df["Annotation"] = normalized_df["Raw Annotation"]
    normalized_df["Year"] = normalized_df["Raw Year"]
    normalized_df["Year-half"] = normalized_df["Raw Year-half"]
    normalized_df["Month"] = normalized_df["Raw Month"]
    normalized_df["Week"] = normalized_df["Raw Week"]
    normalized_df["Client"] = normalized_df["Raw Client"]
    normalized_df["Invoice"] = normalized_df["Raw Invoice"]
    normalized_df["Invoice row ordinal"] = normalized_df["Raw Invoice row ordinal"]
    normalized_df["Invoice row item"] = normalized_df["Raw Invoice row item"]
    normalized_df["Invoice row item category"] = normalized_df[
        "Raw Invoice row item category"
    ]
    normalized_df["Invoice row item reference"] = normalized_df[
        "Raw Invoice row item reference"
    ]
    normalized_df["Invoice hours"] = normalized_df["Raw Invoice hours"]

    normalized_df["Original data"] = df[
        [
            "gmtTimestamp",
            "category",
            "date",
            "dateRaw",
            "hours",
            "hoursRounded",
            "lineWithoutDate",
            "text",
            "ts",
            "tz",
            "sessionMeta.session_ref",
            "sessionMeta.tzFirst",
        ]
    ].to_dict(orient="records")

    # print("normalized_df.columns", normalized_df.columns)

    return normalized_df[return_columns]


def neamtime_tslog_parsing_metadata_to_general_clerk_format(
    parsing_metadata, processing_errors_count
):
    # print("tslog.py - parsing_metadata", parsing_metadata)
    # print("tslog.py - parsing_metadata.columns", parsing_metadata.columns)
    parsing_metadata_row = parsing_metadata.iloc[0]
    # print("tslog.py - parsing_metadata_row", parsing_metadata_row)

    if "troubleshootingInfo.logMetadata.hoursTotal" in parsing_metadata_row:
        """
        Case 1
        """
        normalized_parsing_metadata = {
            "Parse status": [
                "OK"
                if processing_errors_count == 0
                else "%s Error%s"
                % (processing_errors_count, "s" if processing_errors_count > 0 else "")
            ],
            "Tracked Hours": [
                parsing_metadata_row["troubleshootingInfo.logMetadata.hoursTotal"]
            ],
            "Calendar Time in Hours": [
                parsing_metadata_row["troubleshootingInfo.logMetadata.hoursLeadTime"]
            ],
            "Sessions": [parsing_metadata_row["sessionCount"]],
            "Processed lines": [parsing_metadata_row["nonEmptyPreprocessedLinesCount"]],
            "Last modified": ["TODO"],
            "Oldest timestamp": [
                datetime.fromtimestamp(
                    parsing_metadata_row["troubleshootingInfo.logMetadata.startTs"]
                )
            ],
            "Most recent timestamp": [
                datetime.fromtimestamp(
                    parsing_metadata_row["troubleshootingInfo.logMetadata.lastTs"]
                )
            ],
            "Timelog comment": [
                parsing_metadata_row["troubleshootingInfo.logMetadata.name"]
            ],
        }
    else:
        """
        Case 2
        'totalReportedTime', 'sessionCount', 'nonEmptyPreprocessedLinesCount',
           'troubleshootingInfo.logMetadata.error'
        """
        normalized_parsing_metadata = {
            "Parse status": ["Pending (Empty)"],
            "Tracked Hours": [parsing_metadata_row["totalReportedTime"]],
            "Calendar Time in Hours": [None],
            "Sessions": [parsing_metadata_row["sessionCount"]],
            "Processed lines": [parsing_metadata_row["nonEmptyPreprocessedLinesCount"]],
            "Last modified": ["TODO"],
            "Oldest timestamp": [None],
            "Most recent timestamp": [None],
            "Timelog comment": [None],
        }
    return pd.DataFrame(normalized_parsing_metadata)


def neamtime_tslog_processing_errors_to_general_clerk_format(processing_errors):
    # print("tslog.py - processing_errors", processing_errors)
    # print("tslog.py - processing_errors.columns", processing_errors.columns)

    if len(processing_errors) == 0:
        return processing_errors

    normalized_processing_errors = pd.DataFrame()
    normalized_processing_errors["Line number"] = processing_errors["sourceLine"]
    normalized_processing_errors["Date entry"] = processing_errors["dateRaw"]
    normalized_processing_errors["Log entry"] = processing_errors["lineWithComment"]
    normalized_processing_errors["Error log"] = processing_errors["log"]
    """
    'date', 'dateRaw', 'formattedUtcDate',
       'lastInterpretTsAndDateErrorMessage', 'lastKnownTimeZone',
       'lastParseLogCommentErrorMessage', 'lastSetTsAndDateErrorClass',
       'lastSetTsAndDateErrorMessage', 'lastUsedTimeZone', 'line',
       'lineWithComment', 'log', 'preprocessedContentsSourceLineIndex',
       'rowsWithTimeMarkersHandled', 'sourceLine', 'ts',
       'parseLogCommentDetectTimeStampMetadata.log',
       'parseLogCommentDetectTimeStampMetadata.lastKnownsBeforeDetectTimeStamp.lastKnownDate',
       'parseLogCommentDetectTimeStampMetadata.lastKnownsBeforeDetectTimeStamp.lastKnownTimeZone',
       'parseLogCommentDetectTimeStampMetadata.lastKnownsBeforeDetectTimeStamp.lastUsedTimeZone',
       'parseLogCommentDetectTimeStampMetadata.dateRawFormat',
       'parseLogCommentDetectTimeStampMetadata.timeZoneRaw',
       'parseLogCommentDetectTimeStampMetadata.timeRaw',
       'parseLogCommentDetectTimeStampMetadata.dateRaw'
    """
    return normalized_processing_errors


def neamtime_tslog_time_tracking_entries_parser(time_tracking_file_path):
    # type: (str) -> DataFrame
    df, parsing_metadata, processing_errors = parse_neamtime_tslog_time_tracking_file(
        time_tracking_file_path
    )
    return (
        neamtime_tslog_time_tracking_entries_to_general_clerk_format(df),
        neamtime_tslog_parsing_metadata_to_general_clerk_format(
            parsing_metadata, len(processing_errors)
        ),
        neamtime_tslog_processing_errors_to_general_clerk_format(processing_errors),
    )
